# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional, Tuple

import numpy as np

import nncf
from nncf.experimental.tensor import Tensor
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.experimental.tensor.functions import numeric as fns
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.fake_quantize import calculate_scale_zero_point

ReductionAxes = Tuple[int, ...]

NF4_QUANTILES = np.array(
    [
        -1.0,
        -0.6961928009986877,
        -0.5250730514526367,
        -0.39491748809814453,
        -0.28444138169288635,
        -0.18477343022823334,
        -0.09105003625154495,
        0.0,
        0.07958029955625534,
        0.16093020141124725,
        0.24611230194568634,
        0.33791524171829224,
        0.44070982933044434,
        0.5626170039176941,
        0.7229568362236023,
        1.0,
    ],
    dtype=np.float32,
)
CENTER_OF_NF4_QUANTILES = np.array(
    [
        -0.8480964004993439,
        -0.6106329262256622,
        -0.4599952697753906,
        -0.33967943489551544,
        -0.23460740596055984,
        -0.13791173323988914,
        -0.045525018125772476,
        0.03979014977812767,
        0.1202552504837513,
        0.2035212516784668,
        0.2920137718319893,
        0.3893125355243683,
        0.5016634166240692,
        0.6427869200706482,
        0.8614784181118011,
    ],
    dtype=np.float32,
)


@dataclass
class CompressedWeight:
    """
    Compressed weight and decompression parameters.

    :param tensor: The tensor with compressed weight.
    :param scale: The decompression scale, in practice it is dequantization scale for the INT quantization.
    :param zero_point: The zero-point, it is the value of the compression type corresponding to the value 0
        in the non-compression realm. Applicable for INT quantization.
    """

    tensor: Tensor
    scale: Tensor
    zero_point: Optional[Tensor] = None


def reshape_weight_for_grouped_quantization(
    weight: Tensor, reduction_axes: ReductionAxes, group_size: int
) -> Tuple[Tensor, int]:
    """
    Reshapes weight for group-wise quantization and return a new reduction axis for collecting statistics per group
    dimension. Having weight with shapes [c_out, c_in] and group size = 128, shape of reshaped weight is
    [c_out, c_in // 128, 128].

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param group_size: Number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
    :return: reshaped weight and new reduction axis.
    """
    assert group_size != -1
    if isinstance(reduction_axes, tuple) and len(reduction_axes) == 1:
        reduction_axes = reduction_axes[0]
    if not isinstance(reduction_axes, int):
        raise NotImplementedError(
            f"Group-wise quantization expects a single reduction axis, but given: {reduction_axes}."
        )
    channel_size = weight.shape[reduction_axes]
    if channel_size % group_size != 0:
        raise nncf.ValidationError(f"Channel size {channel_size} should be divisible by size of group {group_size}")

    num_groups_per_channel = channel_size // group_size
    shape = list(weight.shape)  # [a1, r, a2] - "r" refers to number of channels along reduction axis
    shape[reduction_axes : reduction_axes + 1] = (num_groups_per_channel, group_size)
    reshaped_weight = weight.reshape(shape)
    reduction_axes += 1
    return reshaped_weight, reduction_axes


def calculate_nf4_scale(weight: Tensor, reduction_axes: ReductionAxes) -> Tensor:
    """
    Calculates the scale for nf4 quantization.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes along which to reduce (collect) different statistics (e.g., min, max).
    :return: Scale tensor of float32 type for nf4 quantization.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    scale = fns.max(fns.abs(weight), axis=reduction_axes, keepdims=True)

    # NOTE: adding machine epsilon to avoid division by zero
    eps = fns.finfo(weight).eps
    scale = fns.where(fns.abs(scale) < eps, eps, scale)

    return scale


def calculate_normalized_weight(weight: Tensor, scale: Tensor) -> Tensor:
    """
    Normalizes the weight tensor using the provided scale.

    :param weight: Weight tensor to normalize.
    :param scale: Scale tensor used for normalization.
    :return: Normalized weight tensor.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)
    if scale.dtype != TensorDataType.float32:
        scale = scale.astype(TensorDataType.float32)

    return weight / scale


def calculate_nf4_weight(weight: Tensor, scale: Tensor) -> Tensor:
    """
    Quantizes the weight tensor to NF4 format.

    :param weight: Weight tensor to quantize.
    :param scale: Scale tensor used for normalization.
    :return: Quantized weight tensor in NF4 format.
    """
    norm_weight = calculate_normalized_weight(weight, scale)

    center_nf4_quantiles = fns.from_numpy(CENTER_OF_NF4_QUANTILES, backend=norm_weight.backend)
    nf4_quantiles = fns.from_numpy(NF4_QUANTILES, backend=norm_weight.backend)

    index_of_quantile = fns.searchsorted(center_nf4_quantiles, norm_weight)
    nf4_weight = nf4_quantiles[index_of_quantile]

    return nf4_weight


def decompress_nf4_weight(weight: Tensor, scale: Tensor) -> Tensor:
    """
    Decompresses the NF4 quantized weight tensor.

    :param weight: Quantized weight tensor in NF4 format.
    :param scale: Scale tensor used for decompression.
    :return: Decompressed weight tensor.
    """
    return weight * scale


def calculate_normalized_weight_and_nf4_scale(
    weight: Tensor, reduction_axes: ReductionAxes, group_size: int = -1, precomputed_scale: Tensor = None
) -> Tuple[Tensor, Tensor]:
    """
    Calculates scale for nf4 quantization and normalizes weights by the scale.
    Weights are reshaped in case of positive value of group size.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param group_size: Number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
        The value -1 means no grouping. Defaults to -1.
    :param precomputed_scale: Precomputed scale.
    :return: Normalized weight tensor of float32 type and nf4 scale tensor of float32 type.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    if group_size != -1:
        # weights are reshaped: [a1, r, a2] -> [a1, r//gs, gs, a2]
        weight, reduction_axes = reshape_weight_for_grouped_quantization(weight, reduction_axes, group_size)

    scale = calculate_nf4_scale(weight, reduction_axes) if precomputed_scale is None else precomputed_scale
    norm_weight = calculate_normalized_weight(weight, scale)
    return norm_weight, scale


def calculate_integer_quantization_params(
    weight: Tensor, reduction_axes: ReductionAxes, config: WeightCompressionConfig
) -> Tuple[Tensor, Tensor]:
    """
    Calculates the scale and zero point for integer quantization.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Weight compression configuration.
    :return: Scale and zero point tensors.
    """
    mode = config.mode
    assert mode != CompressWeightsMode.NF4, "The function supports integer quantization only"
    num_bits = config.num_bits

    level_low = 0
    level_high = 2**num_bits - 1

    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    if mode in [CompressWeightsMode.INT8_ASYM, CompressWeightsMode.INT4_ASYM]:
        min_values = fns.min(weight, axis=reduction_axes, keepdims=True)  # [a1, r, a2] -> [a1, 1, a2]
        max_values = fns.max(weight, axis=reduction_axes, keepdims=True)  # [a1, r, a2] -> [a1, 1, a2]
        scale, zero_point = calculate_scale_zero_point(
            min_values, max_values, level_low, level_high, narrow_range=False
        )
    else:
        level_low_sym = -(2 ** (num_bits - 1))
        level_high_sym = 2 ** (num_bits - 1) - 1

        scale = fns.max(fns.abs(weight), axis=reduction_axes, keepdims=True)  # [a1, r//gs, 1, a2]
        scale = scale / level_high_sym
        zero_point = fns.as_tensor_like(scale, [-level_low_sym]).astype(TensorDataType.int32)
        eps = fns.finfo(scale).eps
        # NOTE: adding machine epsilon to avoid division by zero
        scale = fns.where(fns.abs(scale) < eps, eps, scale)

    return scale, zero_point


def calculate_quantized_weight(
    weight: Tensor, scale: Tensor, zero_point: Tensor, config: WeightCompressionConfig
) -> Tensor:
    """
    Quantizes the weight tensor using the provided scale and zero point.

    :param weight: Weight tensor to quantize.
    :param scale: Scale tensor used for quantization.
    :param zero_point: Zero point tensor used for quantization.
    :param config: Weight compression configuration.
    :return: Quantized weight tensor of uint8 type.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)
    if scale.dtype != TensorDataType.float32:
        scale = scale.astype(TensorDataType.float32)

    num_bits = config.num_bits

    level_low = 0
    level_high = 2**num_bits - 1

    compressed_weights = fns.round(weight / scale + zero_point.astype(weight.dtype))
    compressed_weights = fns.clip(compressed_weights, level_low, level_high).astype(TensorDataType.uint8)
    return compressed_weights


def do_integer_quantization(
    weight: Tensor,
    reduction_axes: ReductionAxes,
    config: WeightCompressionConfig,
    precomputed_scale: Tensor = None,
    precomputed_zero_point: Tensor = None,
) -> Tuple[Tensor, Tensor, Tensor]:
    """
    The method quantizes the given weights to integer data type in accordance with the compression config.
    The config defines a quantization mode:
        INT8_SYM mode refers to unsigned int8 symmetric weight compression with a fixed zero point equals to 128 -
            quantization to [0, 255] range.
        INT8_ASYM mode refers to unsigned int8 asymmetric weight compression with a typical non-fixed zero-point -
            quantization to [0, 255] range.
        INT4_ASYM mode refers to unsigned int4 asymmetric weight compression with a typical non-fixed zero-point -
            quantization to [0, 15] range.
        INT4_SYM mode refers to unsigned int4 symmetric weight compression with a fixed zero point equals to 8 -
            quantization to [0, 15] range.
        NF4 mode requires a dedicated procedure and it is not supported in this method.
    One of the parameter of compression config is a group size. Quantization is per-channel, if group size equals to -1,
    otherwise it's per-group, i.e. group size number of weights in the channel dimension share quantization parameters
    (scales).

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Information on how to compress (quantize) a specific weight.
    :param precomputed_scale: Precomputed scale.
    :param precomputed_zero_point: Precomputed zero point.
    :return: The compressed weights tensor of uint8 type, scale tensor of float32 type and
        zero point tensor of int32 type that was used for its quantization.
    """
    mode = config.mode
    assert mode != CompressWeightsMode.NF4, "The function supports integer quantization only"
    group_size = config.group_size

    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    if group_size != -1:
        # weights are reshaped from [a1, r, a2] to [a1, r//gs, gs, a2]
        weight, reduction_axes = reshape_weight_for_grouped_quantization(weight, reduction_axes, group_size)

    if precomputed_zero_point is None or precomputed_zero_point is None:
        scale, zero_point = calculate_integer_quantization_params(weight, reduction_axes, config)
    if precomputed_scale is not None:
        scale = precomputed_scale
    if precomputed_zero_point is not None:
        zero_point = precomputed_zero_point

    compressed_weights = calculate_quantized_weight(weight, scale, zero_point, config)
    return compressed_weights, scale, zero_point


def get_integer_quantization_error(
    weight: Tensor, reduction_axes: ReductionAxes, config: WeightCompressionConfig
) -> float:
    """
    Calculates a quantity characterizing the difference between floating point weights and fake quantized
    (compressed and decompressed) to integer ones.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Information on how to compress (quantize) a specific weight.
    :return: The quantity characterizing the error of integer quantization.
    """
    orig_shape = weight.shape

    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    compressed_weights, scale, zero_point = do_integer_quantization(weight, reduction_axes, config)

    decompressed_weight = (compressed_weights - zero_point).astype(weight.dtype) * scale

    decompressed_weight = decompressed_weight.reshape(orig_shape)
    diff = (decompressed_weight - weight) ** 2
    layer_err = fns.mean(diff, axis=reduction_axes)
    val = fns.max(layer_err)
    return val.item()


def compress_weight(
    weight: Tensor,
    reduction_axes: ReductionAxes,
    config: WeightCompressionConfig,
    precomputed_scale: Tensor = None,
    precomputed_zero_point: Tensor = None,
):
    """
    Compress weight using compression configuration.

    :param weight: The weight to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Compression configuration.
    :param precomputed_scale: Precomputed scale.
    :param precomputed_zero_point: Precomputed zero point.
    :return: The compressed weight and decompression parameters as instance of CompressedWeight
    """
    if config.mode == CompressWeightsMode.NF4:
        compressed_weight, scale = calculate_normalized_weight_and_nf4_scale(
            weight, reduction_axes, config.group_size, precomputed_scale
        )
        return CompressedWeight(compressed_weight, scale)
    compressed_weight, scale, zero_point = do_integer_quantization(
        weight, reduction_axes, config, precomputed_scale, precomputed_zero_point
    )

    return CompressedWeight(compressed_weight, scale, zero_point)


def do_dequantization(
    compressed_weights: Tensor, scale: Tensor, zero_point: Tensor, reduction_axis: int = -1
) -> Tensor:
    """
    The method dequantizes the given weights to float point data type in accordance with the scale and
    zero_point data type.

    :param compressed_weights: compressed weights.
    :param scale: scale in compression/quantization.
    :param zero_point: zero point in compression/quantization.
    :param reduction_axis: axis for return back for group compression.
    :return: dequantized/decompressed weights.
    """
    decompressed_weight = compressed_weights.astype(dtype=scale.dtype)
    decompressed_weight = (decompressed_weight - zero_point) * scale

    if reduction_axis > -1:
        shape = list(decompressed_weight.shape)  # [a1, r, a2] - "r" refers to number of channels along reduction axis
        shape[reduction_axis] = shape[reduction_axis] * shape[reduction_axis + 1]
        shape[reduction_axis + 1] = 1
        reshaped_weight = decompressed_weight.reshape(shape)
        reshaped_weight = fns.squeeze(reshaped_weight)
        decompressed_weight = reshaped_weight

    return decompressed_weight



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

import tensorflow as tf

from nncf.common.quantization.collectors import QuantizationStatisticsCollector
from nncf.common.quantization.collectors import QuantizerDescription
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.tensorflow.graph.utils import get_nncf_operations
from nncf.tensorflow.quantization.utils import collect_fake_quantize_layers


class TFQuantizationStatisticsCollector(QuantizationStatisticsCollector):
    """
    Implementation of the quantization statistics collector for the TensorFlow backend.
    """

    def __init__(self, model: tf.keras.Model, operation_names: List[str]):
        """
        Initializes a collector of the quantization statistics.

        :param model: An instance of the `tf.keras.Model` class.
        :param operation_names: List of names of quantizer operations. It includes
            names of weight and non-weight quantizers.
        """
        self._model = model
        self._operation_names = operation_names

    def _collect_quantizers_descriptions(self) -> List[QuantizerDescription]:
        """
        Collects descriptions of the quantizers.

        :return: Descriptions of the quantizers.
        """
        quantizers_descriptions = []

        for wrapped_layer, _, op in get_nncf_operations(self._model, self._operation_names):
            is_symmetric = op.mode == QuantizationMode.SYMMETRIC

            is_signed = True
            if is_symmetric:
                operation_weights = wrapped_layer.get_operation_weights(op.name)
                is_signed = op.signed(operation_weights)

            quantizers_descriptions.append(
                QuantizerDescription(op.num_bits, op.per_channel, is_signed, is_symmetric, True, op.enabled)
            )

        for fq_layer in collect_fake_quantize_layers(self._model):
            is_symmetric = fq_layer.mode == QuantizationMode.SYMMETRIC

            quantizers_descriptions.append(
                QuantizerDescription(
                    fq_layer.num_bits, fq_layer.per_channel, fq_layer.signed, is_symmetric, False, fq_layer.enabled
                )
            )

        return quantizers_descriptions

    def _get_potential_quantizers_num(self) -> Tuple[int, int]:
        """
        Returns a potential number of quantizers for weights and activations.

        :return: A tuple (wq_potential_num, aq_potential_num) where
            - `wq_potential_num` is a potential number of quantizers for weights.
            - `aq_potential_num` is a potential number of quantizers for activations.
        """
        return None, None



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any, Dict

import tensorflow as tf

from nncf.common.compression import BaseCompressionAlgorithmController

# TODO(achurkin): remove pylint ignore after 120296 ticked is fixed


class TFCompressionState(tf.train.experimental.PythonState):
    """
    A wrapper for `BaseCompressionAlgorithmController` that allows saving
    the compression state to the checkpoint.
    """

    def __init__(self, controller: BaseCompressionAlgorithmController):
        """
        Initializes the wrapper for the controller.

        :param controller: The controller which gives the compressions state.
        """
        self._ctrl = controller

    def serialize(self) -> str:
        """
        Callback to serialize the compression state.

        :return: A serialized compression state.
        """
        compression_state = self._ctrl.get_compression_state()
        return json.dumps(compression_state)

    def deserialize(self, string_value: str) -> None:
        """
        Callback to deserialize the compression state.

        :param string_value: A serialized compression state.
        """
        compression_state = json.loads(string_value)
        ctrl_state = compression_state[BaseCompressionAlgorithmController.CONTROLLER_STATE]
        self._ctrl.load_state(ctrl_state)


class TFCompressionStateLoader(tf.train.experimental.PythonState):
    """
    This is a class that allows extracting of the compression state from a checkpoint.
    The extracted compression state is not applied.
    """

    def __init__(self):
        """
        Initializes the compression state loader.
        """
        self._state = None

    @property
    def state(self) -> Dict[str, Any]:
        """
        Returns the compression state which was extracted from the checkpoint.

        :return: The compression state.
        """
        return self._state

    def serialize(self) -> str:
        raise NotImplementedError(
            "Use an instance of the `TFCompressionState` class to serialize the compression state."
        )

    def deserialize(self, string_value: str) -> None:
        """
        Callback to deserialize the compression state.

        :param string_value: A serialized compression state.
        """
        self._state = json.loads(string_value)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Callable, Dict, Iterable, List, Optional, Tuple, TypeVar

import numpy as np
import torch

import nncf
from nncf import nncf_logger
from nncf.common.graph.layer_attributes import Dtype
from nncf.torch.dynamic_graph.context import TracingContext
from nncf.torch.dynamic_graph.graph import DynamicGraphNode
from nncf.torch.dynamic_graph.op_input_processing import OperatorInput
from nncf.torch.dynamic_graph.trace_tensor import TensorMeta
from nncf.torch.dynamic_graph.trace_tensor import TracedTensor
from nncf.torch.dynamic_graph.trace_tensor import TracedTensorMixin
from nncf.torch.nested_objects_traversal import objwalk

TensorOrTupleOrList = TypeVar("TensorOrTupleOrList", List[torch.Tensor], Tuple[torch.Tensor], torch.Tensor)


def is_iterable(item):
    non_iterable_types = (str, bytes, bytearray, torch.Tensor, np.ndarray)

    return isinstance(item, Iterable) and not isinstance(item, non_iterable_types)


def flatten(items):
    it = items.items() if hasattr(items, "items") else iter(items)
    for item in it:
        if is_iterable(item):
            for i in flatten(item):
                yield i
        else:
            yield item


def flatten_args(args, kwargs):
    return list(flatten(args)) + list(flatten(kwargs))


def get_dtype(x: torch.Tensor) -> Dtype:
    if x.dtype in [torch.float, torch.float16, torch.bfloat16, torch.float32, torch.float64]:
        return Dtype.FLOAT
    return Dtype.INTEGER


def forward_trace_only(operator: Callable, *args, **kwargs):
    """
    This wrapper override will result in the operator not being added to graph,
    but the result will still have TracedTensors with parent IDs left the same as in input.
    Useful for operators which are not likely to be present in patterns considered for
    compression, but still have to be accounted for so that the NNCF internal graph representation
    does not become disjoint.
    """

    result = operator(*args, **kwargs)

    fargs = flatten_args(args, kwargs)
    input_traced_tensor_indices = [i for i in range(len(fargs)) if isinstance(fargs[i], TracedTensor)]

    if isinstance(result, (list, tuple)):
        output_tensors_to_be_traced_indices = [i for i in range(len(result)) if isinstance(result[i], torch.Tensor)]

        was_tuple = isinstance(result, tuple)
        result = list(result)
        if len(input_traced_tensor_indices) == 1:
            # Broadcast one and the same creator ID of input to all outputs
            for out_idx in output_tensors_to_be_traced_indices:
                forwarded_meta = deepcopy(fargs[input_traced_tensor_indices[0]].tensor_meta)
                if forwarded_meta is not None:
                    forwarded_meta.shape = tuple(result[out_idx].shape)
                result[out_idx] = TracedTensor.from_torch_tensor(result[out_idx], forwarded_meta)
        elif len(input_traced_tensor_indices) != len(output_tensors_to_be_traced_indices):
            raise nncf.ValidationError(
                "Unable to forward trace through operator {} - "
                "input and output tensor count mismatch!".format(operator.__name__)
            )
        else:
            # Assume that output tensor order corresponds to input tensor order
            for in_idx, out_idx in zip(input_traced_tensor_indices, output_tensors_to_be_traced_indices):
                forwarded_meta = deepcopy(fargs[in_idx].tensor_meta)
                if forwarded_meta is not None:
                    forwarded_meta.shape = tuple(result[out_idx].shape)
                result[out_idx] = TracedTensor.from_torch_tensor(result[out_idx], forwarded_meta)
        if was_tuple:
            result = tuple(result)
    elif len(input_traced_tensor_indices) > 1:
        raise nncf.ValidationError(
            "Unable to forward trace through operator {} - "
            "input and output tensor count mismatch!".format(operator.__name__)
        )
    elif input_traced_tensor_indices:
        forwarded_meta = deepcopy(fargs[input_traced_tensor_indices[0]].tensor_meta)
        if forwarded_meta is not None:
            forwarded_meta.shape = tuple(result.shape)
        return TracedTensor.from_torch_tensor(result, forwarded_meta)
    # No traced tensors in input, return a usual torch.Tensor as well
    return result


def trace_tensor(x: torch.Tensor, port_id: int, node: DynamicGraphNode, ctx: TracingContext) -> torch.Tensor:
    """
    Dynamically turn torch.Tensor instance into tensor with tracing capabilities instance using TracedTensorMixin.

    :param x: The input Tensor.
    :param node: A node in DynamicGraph associated with the function that produced the input tensor
    :param ctx: The resulting tensors will be registered within this TracingContext instance
        to be stipped on context exit, which is required to correctly process situations when a traced model
        retains intermediate tensor values.
    :return: The resulting tensor with the TracedTensorMixin interface.
    """
    if not isinstance(x, torch.Tensor):
        nncf_logger.debug(f"Could not find tensors to trace in operator output: {x}")
        return x

    meta = None
    if node is not None:
        meta = TensorMeta(node.node_id, port_id, x.shape, get_dtype(x))

    if isinstance(x, TracedTensorMixin):
        tt = x
        tt.tensor_meta = meta
    else:
        tt = TracedTensor.from_torch_tensor(x, meta)
    ctx.register_traced_tensor(tt)
    return tt


def trace_tensors(
    operator_output: TensorOrTupleOrList, node: "DynamicGraphNode", ctx: "TracingContext"
) -> TensorOrTupleOrList:
    """
    Dynamically turn torch.Tensor instances in `operator_output` into tensor with tracing capabilities instances
    using TracedTensorMixin. `operator_output` is presumed to be the output of a model operation (function call)
    associated with `node`.

    :param operator_output: The output of an NNCF-wrapped function executed in a model object.
    :param node: A node in DynamicGraph associated with the function that produced `operator_output`
    :param ctx: The resulting tensors will be registered within this TracingContext instance
        to be stipped on context exit, which is required to correctly process situations when a traced model
        retains intermediate tensor values.
    :return: Same structure as `operator_output`, but with torch.Tensor entries turned into tensor
        with tracing capabilities instance using TracedTensorMixin.
    """

    if isinstance(operator_output, (list, tuple)):
        output_ = []
        for i, x in enumerate(operator_output):
            tt = trace_tensor(x, i, node, ctx)
            output_.append(tt)
        return operator_output.__class__(output_)
    return trace_tensor(operator_output, 0, node, ctx)


def make_tensor_metas(inputs: OperatorInput) -> List[Optional[TensorMeta]]:
    """
    Produces TensorMeta data for each torch.Tensor or TracedTensorMixin in `inputs`.

    :param inputs: An OperatorInput representation of input arguments to an operation in the traced model.
    :return: A list of TensorMeta objects, one for every torch.Tensor or TracedTensorMixin object in `inputs` in the
    order of item enumeration in `inputs`.
    """
    tensor_metas = []
    for i, node_input_index_entry in enumerate(inputs):
        node_input = node_input_index_entry.getter()
        if isinstance(node_input, TracedTensorMixin):
            tensor_metas.append(node_input.tensor_meta)
        elif isinstance(node_input, torch.Tensor):
            meta = TensorMeta(None, i, node_input.shape, get_dtype(node_input))
            tensor_metas.append(meta)
        else:
            tensor_metas.append(None)
    return tensor_metas


def strip_traced_tensors(args: Tuple, kwargs: Dict) -> Tuple[Tuple, Dict]:
    """
    Required to guard against new forward calls on tensors that have already passed
    through NNCF's forward once and got turned into TracedTensors by reference access.
    """
    is_traced_tensor_predicate = lambda x: isinstance(x, TracedTensorMixin)
    strip_traced_tensor = lambda x: x.strip()

    args = objwalk(args, is_traced_tensor_predicate, strip_traced_tensor)
    kwargs = objwalk(kwargs, is_traced_tensor_predicate, strip_traced_tensor)
    return args, kwargs



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Tuple, Union

import torch

from nncf.experimental.tensor.functions import linalg


@linalg.norm.register(torch.Tensor)
def _(
    a: torch.Tensor,
    ord: Optional[Union[str, float, int]] = None,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdims: bool = False,
) -> torch.Tensor:
    return torch.linalg.norm(a, ord=ord, dim=axis, keepdims=keepdims)


@linalg.cholesky.register(torch.Tensor)
def _(a: torch.Tensor, upper: bool = False) -> torch.Tensor:
    return torch.linalg.cholesky(a, upper=upper)


@linalg.cholesky_inverse.register(torch.Tensor)
def _(a: torch.Tensor, upper: bool = False) -> torch.Tensor:
    return torch.cholesky_inverse(a, upper=upper)


@linalg.inv.register(torch.Tensor)
def _(a: torch.Tensor) -> torch.Tensor:
    return torch.linalg.inv(a)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import warnings
from typing import Callable, Type, TypeVar

from packaging import version


def warning_deprecated(msg):
    # Note: must use FutureWarning in order not to get suppressed by default
    warnings.warn(msg, FutureWarning, stacklevel=2)


ClassOrFn = TypeVar("ClassOrFn", Callable, Type)


class deprecated:
    """
    A decorator for marking function calls or class instantiations as deprecated. A call to the marked function or an
    instantiation of an object of the marked class will trigger a `FutureWarning`. If a class is marked as
    @deprecated, only the instantiations will trigger a warning, but static attribute accesses or method calls will not.
    """

    def __init__(self, msg: str = None, start_version: str = None, end_version: str = None):
        """
        :param msg: Custom message to be added after the boilerplate deprecation text.
        """
        self.msg = msg
        self.start_version = version.parse(start_version) if start_version is not None else None
        self.end_version = version.parse(end_version) if end_version is not None else None

    def __call__(self, fn_or_class: ClassOrFn) -> ClassOrFn:
        name = fn_or_class.__module__ + "." + fn_or_class.__name__
        if inspect.isclass(fn_or_class):
            fn_or_class.__init__ = self._get_wrapper(fn_or_class.__init__, name)
            return fn_or_class
        return self._get_wrapper(fn_or_class, name)

    def _get_wrapper(self, fn_to_wrap: Callable, name: str) -> Callable:
        @functools.wraps(fn_to_wrap)
        def wrapped(*args, **kwargs):
            msg = f"Usage of {name} is deprecated "
            if self.start_version is not None:
                msg += f"starting from NNCF v{str(self.start_version)} "
            msg += "and will be removed in "
            if self.end_version is not None:
                msg += f"NNCF v{str(self.end_version)}."
            else:
                msg += "a future NNCF version."
            if self.msg is not None:
                msg += "\n" + self.msg
            warning_deprecated(msg)
            return fn_to_wrap(*args, **kwargs)

        return wrapped



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from collections import defaultdict
from copy import deepcopy
from typing import Any, Callable, Dict, Generator, KeysView, List, Optional, Tuple, Type, ValuesView, cast

import networkx as nx  # type:ignore
import networkx.algorithms.isomorphism as iso  # type:ignore
from networkx.classes.reportviews import OutEdgeView  # type:ignore

import nncf
from nncf.common.graph.graph_matching import find_subgraphs_matching_pattern
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.patterns import GraphPattern
from nncf.common.utils.dot_file_rw import write_dot_graph

NNCFNodeName = str
LayerName = str


class NNCFNode:
    """
    Class describing nodes used in NNCFGraph.
    """

    ID_NODE_ATTR = "id"
    NODE_NAME_ATTR = "node_name"
    KEY_NODE_ATTR = "key"
    NODE_TYPE_ATTR = "type"
    METATYPE_ATTR = "metatype"
    LAYER_NAME_ATTR = "layer_name"
    LAYER_ATTRIBUTES = "layer_attributes"
    IGNORED_ALGOS_ATTR = "ignored_algos"
    IS_IN_ITERATION_SCOPE_NODE_ATTR = "is_in_iteration_scope"
    IS_INTEGER_INPUT_NODE_ATTR = "is_integer_input"
    IS_SHARED_ATTR = "is_shared"

    def __init__(self, attributes: Dict[str, Any]) -> None:
        self._attributes = attributes

    @property
    def attributes(self) -> Dict[str, Any]:
        return self._attributes

    @property
    def node_id(self) -> int:
        return cast(int, self._attributes[NNCFNode.ID_NODE_ATTR])

    @property
    def node_key(self) -> str:
        return cast(str, self._attributes[NNCFNode.KEY_NODE_ATTR])

    @property
    def node_name(self) -> NNCFNodeName:
        return cast(NNCFNodeName, self._attributes[NNCFNode.NODE_NAME_ATTR])

    @property
    def metatype(self) -> Type[OperatorMetatype]:
        return cast(Type[OperatorMetatype], self._attributes[NNCFNode.METATYPE_ATTR])

    @property
    def node_type(self) -> str:
        return cast(str, self._attributes[NNCFNode.NODE_TYPE_ATTR])

    @property
    def layer_name(self) -> Optional[LayerName]:
        return self._attributes.get(NNCFNode.LAYER_NAME_ATTR)

    @layer_name.setter
    def layer_name(self, value: str) -> None:
        self._attributes[NNCFNode.LAYER_NAME_ATTR] = value

    @property
    def layer_attributes(self) -> Optional[BaseLayerAttributes]:
        return self._attributes.get(NNCFNode.LAYER_ATTRIBUTES)

    @layer_attributes.setter
    def layer_attributes(self, value: BaseLayerAttributes) -> None:
        self._attributes[NNCFNode.LAYER_ATTRIBUTES] = value

    @property
    def ignored_algorithms(self) -> List[str]:
        return cast(List[str], self._attributes[NNCFNode.IGNORED_ALGOS_ATTR])

    def is_in_iteration_scope(self) -> bool:
        return cast(bool, self._attributes[NNCFNode.IS_IN_ITERATION_SCOPE_NODE_ATTR])

    def is_integer_input(self) -> bool:
        return cast(bool, self._attributes[NNCFNode.IS_INTEGER_INPUT_NODE_ATTR])

    def is_shared(self) -> bool:
        return cast(bool, self._attributes[NNCFNode.IS_SHARED_ATTR])

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return " ".join([str(self.node_id), self.node_name, self.node_type])

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NNCFNode) and self.attributes == other.attributes


class NNCFGraphEdge:
    """
    A structure describing an edge in NNCFGraph. Since nodes of the NNCFGraph are operations
    on (activation) tensors, an edge in NNCFGraph is a representation of an activation tensor produced or
    consumed by an operation.
    """

    def __init__(
        self,
        from_node: NNCFNode,
        to_node: NNCFNode,
        input_port_id: int,
        output_port_id: int,
        tensor_shape: List[int],
        dtype: Dtype,
        parallel_input_port_ids: List[int],
    ) -> None:
        """
        :param from_node: An NNCFNode that sources the directed edge.
        :param to_node: An NNCFNode that sinks the directed edge.
        :param input_port_id: The ID of the tensor input to the `to_node` that this edge corresponds to.
        :param output_port_id: The ID of the tensor output of the `from_node` that this edge corresponds to..
        :param tensor_shape: The shape of the activation tensor the edge represents.
        :param dtype: The data type of the activation tensor the edge represents.
        """
        self.from_node = from_node
        self.to_node = to_node
        self.input_port_id = input_port_id
        self.output_port_id = output_port_id
        self.tensor_shape: Tuple[int, ...] = tuple(tensor_shape)
        self.dtype = dtype
        self.parallel_input_port_ids = parallel_input_port_ids

    def __str__(self) -> str:
        return f"{self.from_node}:{self.output_port_id} -> {self.tensor_shape} -> {self.to_node}:{self.input_port_id}"

    def __hash__(self) -> int:
        return hash(
            (
                self.from_node,
                self.to_node,
                self.input_port_id,
                self.output_port_id,
                tuple(self.tensor_shape),
                self.dtype,
                tuple(self.parallel_input_port_ids),
            )
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NNCFGraphEdge) and self.__dict__ == other.__dict__


class NNCFGraphPatternIO:
    """
    Describes the inputs and outputs of a subgraph in NNCFGraph.
    """

    def __init__(self, input_edges: List[NNCFGraphEdge], output_edges: List[NNCFGraphEdge]):
        self.input_edges = input_edges
        self.output_edges = output_edges


class NNCFGraph:
    """
    Wrapper over a regular directed acyclic graph that represents a control flow/execution graph of a DNN
    providing some useful methods for graph traversal.
    """

    ACTIVATION_SHAPE_EDGE_ATTR = "activation_shape"
    INPUT_PORT_ID_EDGE_ATTR = "input_port_id"
    OUTPUT_PORT_ID_EDGE_ATTR = "output_port_id"
    DTYPE_EDGE_ATTR = "dtype"
    PARALLEL_INPUT_PORT_IDS_ATTR = "parallel_input_ports"

    def __init__(self) -> None:
        self._nx_graph = nx.DiGraph()
        self._node_id_to_key_dict: Dict[int, str] = {}
        self._nodes: Dict[str, NNCFNode] = {}
        self._input_nncf_nodes: Dict[int, NNCFNode] = {}
        self._output_nncf_nodes: Dict[int, NNCFNode] = {}
        self._node_ids_vs_layer_names: Dict[int, LayerName] = {}
        self._layer_name_vs_shared_nodes: Dict[LayerName, List[NNCFNode]] = defaultdict(list)
        self._node_name_to_node_id_map: Dict[str, List[int]] = {}

    @property
    def nodes(self) -> Dict[str, NNCFNode]:
        return self._nodes

    def get_node_by_id(self, node_id: int) -> NNCFNode:
        """
        :param node_id: Id of the node.
        :return: Node in a graph with such id.
        """
        return self.get_node_by_key(self.get_node_key_by_id(node_id))

    def get_node_by_key(self, key: str) -> NNCFNode:
        """
        :param key: key (node_name) of the node.
        :return: NNCFNode in a graph with such key.
        """
        return self._nodes[key]

    def get_input_nodes(self) -> List[NNCFNode]:
        """
        :return: List of input nodes of the graph.
        """
        return list(self._input_nncf_nodes.values())

    def get_output_nodes(self) -> List[NNCFNode]:
        """
        :return: List of output nodes of the graph.
        """
        return list(self._output_nncf_nodes.values())

    def get_nodes_by_types(self, type_list: List[str]) -> List[NNCFNode]:
        """
        :param type_list: List of types to look for.
        :return: List of nodes with provided types.
        """
        all_nodes_of_type = []
        for nncf_node in self.get_all_nodes():
            if nncf_node.node_type in type_list:
                all_nodes_of_type.append(nncf_node)
        return all_nodes_of_type

    def get_nodes_by_metatypes(self, metatype_list: List[Type[OperatorMetatype]]) -> List[NNCFNode]:
        """
        Return a list of nodes with provided metatypes.

        :param metatype_list: List of types to look for.
        :return: List of nodes with provided metatypes.
        """
        all_nodes_of_type = []
        for nncf_node in self.get_all_nodes():
            if nncf_node.metatype in metatype_list:
                all_nodes_of_type.append(nncf_node)
        return all_nodes_of_type

    def get_all_node_ids(self) -> KeysView[int]:
        """
        Returns all graph nodes' node_ids.
        """
        return self._node_id_to_key_dict.keys()

    def get_all_node_keys(self) -> ValuesView[str]:
        """
        Returns all graph nodes' keys i.e. node_names.
        """
        return self._node_id_to_key_dict.copy().values()

    def get_all_nodes(self) -> List[NNCFNode]:
        """
        Returns list of all graph nodes.
        """
        return list(self._nodes.values())

    def get_all_simple_paths(
        self, start_node_name: NNCFNodeName, end_node_name: NNCFNodeName
    ) -> Generator[List[NNCFNodeName], None, None]:
        """
        Generates all simple paths in the NNCFGraph from start node to end node.
        A simple path is a path with no repeated nodes.

        :param start_node_name: a name of starting node for path
        :param end_node_name: a name of node at which to end path
        :return: A generator that produces lists of simple paths. If there are no paths between the start and end nodes
        the generator produces no output.
        """
        start_node = self.get_node_by_name(start_node_name)
        end_node = self.get_node_by_name(end_node_name)
        start_node_key = self.get_node_key_by_id(start_node.node_id)
        end_node_key = self.get_node_key_by_id(end_node.node_id)
        return cast(
            Generator[List[NNCFNodeName], None, None], nx.all_simple_paths(self._nx_graph, start_node_key, end_node_key)
        )

    @staticmethod
    def _get_edge_boundaries(
        match: List[str], graph: nx.DiGraph
    ) -> Tuple[List[Tuple[str, str, Dict[str, Any]]], List[Tuple[str, str, Dict[str, Any]]]]:
        out_edge_boundary = list(nx.edge_boundary(graph, match, data=True))
        complement = list(filter(lambda x: x not in match, graph.nodes.keys()))
        in_edge_boundary = list(nx.edge_boundary(graph, complement, data=True))
        return sorted(in_edge_boundary), sorted(out_edge_boundary)  # must be sorted for determinism

    def get_node_key_by_id(self, node_id: int) -> str:
        """
        Returns node key (node_name) by provided id.

        :param node_id: Id of the node.
        :return: Key of the node with provided id.
        """
        return self._node_id_to_key_dict[node_id]

    def get_next_nodes(self, node: NNCFNode) -> List[NNCFNode]:
        """
        Returns consumer nodes of provided node.

        :param node: Producer node.
        :return: List of consumer nodes of provided node.
        """
        nx_node_keys = self._nx_graph.succ[self._node_id_to_key_dict[node.node_id]]
        return [self._nodes[key] for key in nx_node_keys]

    def get_previous_nodes(self, node: NNCFNode) -> List[NNCFNode]:
        """
        Returns producer nodes of provided node.

        :param node: Consumer node.
        :return: List of producers nodes of provided node.
        """

        nx_node_keys = self._nx_graph.pred[self._node_id_to_key_dict[node.node_id]]
        return [self._nodes[key] for key in nx_node_keys]

    def get_input_edges(self, node: NNCFNode) -> List[NNCFGraphEdge]:
        """
        Returns edges of input tensors with description sorted by input port ID.

        :param node: Consumer node.
        :return: List of input edges for the node sorted by input port ID.
        """
        input_nodes = self.get_previous_nodes(node)
        edges = []
        for from_node in input_nodes:
            edges.extend(self._get_edges(from_node, node))
        return sorted(edges, key=lambda x: x.input_port_id)

    def get_output_edges(self, node: NNCFNode) -> List[NNCFGraphEdge]:
        """
        Returns edges of output tensors sorted by output port ID.

        :param node: Producer node.
        :return:  List of output edges for the node sorted by output port ID.
        """

        output_nodes = self.get_next_nodes(node)
        edges = []
        for to_node in output_nodes:
            edges.extend(self._get_edges(node, to_node))
        return sorted(edges, key=lambda x: x.output_port_id)

    def _get_edges(self, from_node: NNCFNode, to_node: NNCFNode) -> List[NNCFGraphEdge]:
        edges = []
        edge = self.get_edge(from_node, to_node)
        parallel_input_port_ids = edge.parallel_input_port_ids
        edge.parallel_input_port_ids = []
        edges.append(edge)
        for input_port_id in parallel_input_port_ids:
            edges.append(
                NNCFGraphEdge(
                    from_node=edge.from_node,
                    to_node=edge.to_node,
                    input_port_id=input_port_id,
                    output_port_id=edge.output_port_id,
                    tensor_shape=list(edge.tensor_shape),
                    dtype=edge.dtype,
                    parallel_input_port_ids=[],
                )
            )
        return edges

    def traverse_graph(
        self,
        curr_node: NNCFNode,
        traverse_function: Callable[[NNCFNode, List[Any]], Tuple[bool, List[Any]]],
        traverse_forward: bool = True,
    ) -> List[Any]:
        """
        Traverses graph up or down starting form `curr_node` node.

        :param curr_node: Node from which traversal is started.
        :param traverse_function: Function describing condition of traversal continuation/termination.
        :param traverse_forward: Flag specifying direction of traversal.
        :return:
        """
        output: List[Any] = []
        return self._traverse_graph_recursive_helper(curr_node, traverse_function, output, traverse_forward)

    def _traverse_graph_recursive_helper(
        self,
        curr_node: NNCFNode,
        traverse_function: Callable[[NNCFNode, List[Any]], Tuple[bool, List[Any]]],
        output: List[Any],
        traverse_forward: bool,
    ) -> List[Any]:
        is_finished, output = traverse_function(curr_node, output)
        get_nodes_fn = self.get_next_nodes if traverse_forward else self.get_previous_nodes
        if not is_finished:
            for node in get_nodes_fn(curr_node):
                self._traverse_graph_recursive_helper(node, traverse_function, output, traverse_forward)
        return output

    def add_nncf_node(
        self,
        node_name: str,
        node_type: str,
        node_metatype: Type[OperatorMetatype],
        layer_attributes: Optional[BaseLayerAttributes] = None,
        node_id_override: Optional[int] = None,
        layer_name: Optional[LayerName] = None,
        ignored_algorithms: Optional[List[str]] = None,
        is_in_iteration_scope: bool = False,
        is_integer_input: bool = False,
        is_shared: bool = False,
    ) -> NNCFNode:
        """
        Adds a node into the graph. A node represents an operation being performed on tensors.
        :param node_name: The name of the node to add - will serve as a human-readable and specifiable ID.
        :param node_type: The type of the node - usually a framework-specific string representation of the operation.
        :param node_metatype: The metatype of the node - a framework-abstract definition of what the operation
            actually means.
        :param layer_attributes: Must be passed for operations that, in order to be processed
            during compression, require additional information such as the exact dimension of the weights tensor to be
            considered a "channel" dimension for per-channel quantization, the weight shape itself for sparsity mask
            creation etc.
        :param node_id_override: The numerical ID to be associated with the new node; if unspecified, will
            assign a unique ID.
        :param layer_name: The name of the framework-specific "layer" object that houses the operation represented by
            the node and associated trainable weights, if any.
        :param ignored_algorithms: A list of compression algorithm names (from the same set of strings that are
            specified in the `"algorithm": ...` section of the .json NNCF config or `ptq_quantization`)
            which should ignore this operation.
        :param is_in_iteration_scope: Whether the node to be currently added corresponds to an iteration of an RNN
            cycle (where the number of iterations is determined dynamically based on the RNN input shape).
        :param is_integer_input: Only valid for input nodes - whether the input node corresponds to an integer input.
        :param is_shared: Whether the node corresponds to an operation that accesses the weights that are also accessed
            by another operation (represented by a separate node) in NNCFGraph. Examples would be repeated applications
            of a convolution layer with the same weights at different stages in the network.
        :return: An NNCFNode object representing the newly created node in the graph. The node will still have
            to be connected to the rest of the nodes with edges using the `NNCFGraph.add_edge_between_nncf_nodes`
            method.
        """
        if node_id_override is not None:
            node_id = node_id_override
        else:
            node_ids = list(self.get_all_node_ids())
            if node_ids:
                node_id = max(self.get_all_node_ids()) + 1
            else:
                node_id = 0

        if node_id in self._node_id_to_key_dict:
            raise ValueError(f"NNCF node with id {node_id} is already in the NNCFGraph")

        node_ids = self._node_name_to_node_id_map.setdefault(node_name, [])
        node_ids.append(node_id)

        node_key = f"{node_id} {node_name}"

        self._node_id_to_key_dict[node_id] = node_key
        attrs = {
            NNCFNode.ID_NODE_ATTR: node_id,
            NNCFNode.NODE_NAME_ATTR: node_name,
            NNCFNode.KEY_NODE_ATTR: node_key,
            NNCFNode.NODE_TYPE_ATTR: node_type,
            NNCFNode.LAYER_NAME_ATTR: layer_name,
            NNCFNode.METATYPE_ATTR: node_metatype,
            NNCFNode.IS_SHARED_ATTR: is_shared,
            NNCFNode.IS_IN_ITERATION_SCOPE_NODE_ATTR: is_in_iteration_scope,
            NNCFNode.IS_INTEGER_INPUT_NODE_ATTR: is_integer_input,
        }
        if layer_attributes is not None:
            attrs[NNCFNode.LAYER_ATTRIBUTES] = layer_attributes

        if ignored_algorithms is None:
            ignored_algorithms = []
        attrs[NNCFNode.IGNORED_ALGOS_ATTR] = ignored_algorithms
        self._nx_graph.add_node(node_key, **attrs)

        node = NNCFNode(self._nx_graph.nodes[node_key])
        self._nodes[node_key] = node

        if node.metatype in INPUT_NOOP_METATYPES:
            self._input_nncf_nodes[node_id] = node

        if node.metatype in OUTPUT_NOOP_METATYPES:
            self._output_nncf_nodes[node_id] = node

        if layer_name is not None:
            self._node_ids_vs_layer_names[node.node_id] = layer_name
            self._layer_name_vs_shared_nodes[layer_name].append(node)

        return node

    def add_edge_between_nncf_nodes(
        self,
        from_node_id: int,
        to_node_id: int,
        tensor_shape: List[int],
        input_port_id: int,
        output_port_id: int,
        dtype: Dtype,
        parallel_input_port_ids: Optional[List[int]] = None,
    ) -> None:
        """
        Adds a directed edge between two `NNCFNode`s that are already present in the graph.
        The edge represents an activation tensor, produced or consumed by an operation (which is represented by a node)
        :param from_node_id: The `NNCFNode.node_id` of the node that produces the tensor represented by the edge.
        :param to_node_id: The `NNCFNode.node_id` of the node that consumes the tensor represented by the edge.
        :param tensor_shape: The shape of the tensor represented by the edge.
        :param input_port_id: Specifies the index among the possible inputs of the `to_node_id` node' that this tensor
            should correspond to.
        :param output_port_id: Specifies the index among the possible outputs of the `from_node_id` node' that this
            tensor should correspond to.
        :param dtype: The data type of the tensor.
        :param parallel_input_port_ids: Input ports for parallel edges, if any should be present for this edge.
        """
        from_node_key = self._node_id_to_key_dict[from_node_id]
        to_node_key = self._node_id_to_key_dict[to_node_id]

        err_reason = None

        if from_node_key not in self._nx_graph.nodes:
            err_reason = f"node {from_node_key} not in NNCFGraph"
        if to_node_key not in self._nx_graph.nodes:
            err_reason = f"node {from_node_key} not in NNCFGraph"
        if from_node_id in self._output_nncf_nodes:
            err_reason = "cannot add edges *from* output nodes"
        if to_node_id in self._input_nncf_nodes:
            err_reason = "cannot add edges *to* input nodes"

        if err_reason is not None:
            raise ValueError(f"Cannot add edge from {from_node_key} to {to_node_key} - {err_reason}!")

        attrs = {
            NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR: tensor_shape,
            NNCFGraph.INPUT_PORT_ID_EDGE_ATTR: input_port_id,
            NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR: output_port_id,
            NNCFGraph.DTYPE_EDGE_ATTR: dtype,
            NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR: [] if parallel_input_port_ids is None else parallel_input_port_ids,
        }
        self._nx_graph.add_edge(from_node_key, to_node_key, **attrs)

    def topological_sort(self) -> List[NNCFNode]:
        """
        Returns nodes in topologically sorted order, additionally sorted in ascending node ID order.
        """
        return [
            self._nodes[node_name]
            for node_name in nx.lexicographical_topological_sort(
                self._nx_graph, key=lambda x: self._nx_graph.nodes[x][NNCFNode.ID_NODE_ATTR]
            )
        ]

    def dump_graph(self, path: str) -> None:
        write_dot_graph(self.get_graph_for_structure_analysis(), pathlib.Path(path))

    def visualize_graph(self, path: str) -> None:
        out_graph = self._get_graph_for_visualization()
        write_dot_graph(out_graph, pathlib.Path(path))

    def get_graph_for_structure_analysis(self, extended: bool = False) -> nx.DiGraph:
        """
        Returns the nx.Digraph, which is built based on self._nx_graph.
        The new graph has certain node attributes omitted, compared to the graph stored inside NNCFGraph.
        If the node name consists of a special reserved character, this character will be replaced.

        :param extended: whether the graph edges should have attributes: shape of the tensor and tensor primitive type.
        :return: An nx.DiGraph to be used for structure analysis
        """
        out_graph = nx.DiGraph()
        for node_name, node in self._nx_graph.nodes.items():
            attrs_node = {"id": str(node[NNCFNode.ID_NODE_ATTR]), "type": node[NNCFNode.NODE_TYPE_ATTR]}
            for attr in ["color", "label", "style"]:
                if attr in node:
                    attrs_node[attr] = node[attr]

            out_graph.add_node(node_name, **attrs_node)

        for u, v in self._nx_graph.edges:
            edge = self._nx_graph.edges[u, v]
            attrs_edge = {}
            label = {}
            if edge[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]:
                label["parallel_input_port_ids"] = edge[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]

            if extended:
                if edge[NNCFGraph.DTYPE_EDGE_ATTR] is Dtype.INTEGER:
                    attrs_edge["style"] = "dashed"
                else:
                    attrs_edge["style"] = "solid"
                label["shape"] = edge[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]

            if label:
                if "shape" in label and len(label) == 1:
                    attrs_edge["label"] = label["shape"]
                else:
                    attrs_edge["label"] = ", ".join((f"{k}:{v}" for k, v in label.items()))
            out_graph.add_edge(u, v, **attrs_edge)
        return out_graph

    def _get_graph_for_visualization(self) -> nx.DiGraph:
        """
        :return: A user-friendly graph .dot file, making it easier to debug the network and setup
        ignored/target scopes.
        """
        out_graph = nx.DiGraph()
        for node in self.get_all_nodes():
            attrs_node = {}
            attrs_node["label"] = f"{node.node_id} {node.node_name}"
            node_key = self.get_node_key_by_id(node.node_id)
            out_graph.add_node(node_key, **attrs_node)

        for u, v in self._nx_graph.edges:
            edge = self._nx_graph.edges[u, v]
            if edge[NNCFGraph.DTYPE_EDGE_ATTR] is Dtype.INTEGER:
                style = "dashed"
            else:
                style = "solid"
            edge_label = (
                f"{edge[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]} \\n"
                f"{edge[NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR]} -> {edge[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR]}"
            )
            out_graph.add_edge(u, v, label=edge_label, style=style)

        mapping = {k: v["label"] for k, v in out_graph.nodes.items()}
        out_graph = nx.relabel_nodes(out_graph, mapping)
        for node in out_graph.nodes.values():
            node.pop("label")  # type: ignore

        return out_graph

    def get_node_by_name(self, name: NNCFNodeName) -> NNCFNode:
        node_ids = self._node_name_to_node_id_map.get(name, None)
        if node_ids is None:
            raise nncf.InternalError("Could not find a node {} in NNCFGraph!".format(name))
        if len(node_ids) > 1:
            raise nncf.InternalError(f"More than one node in NNCFGraph matches name {name}")

        node_key = f"{node_ids[0]} {name}"
        return self._nodes[node_key]

    def __eq__(self, other: object) -> bool:
        nm = iso.categorical_node_match(
            [NNCFNode.ID_NODE_ATTR, NNCFNode.KEY_NODE_ATTR, NNCFNode.LAYER_ATTRIBUTES], [None, None, None]
        )
        em = iso.categorical_edge_match(
            [NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR, NNCFGraph.INPUT_PORT_ID_EDGE_ATTR], [None, None]
        )
        return isinstance(other, NNCFGraph) and bool(
            nx.is_isomorphic(self._nx_graph, other._nx_graph, node_match=nm, edge_match=em)
        )

    def get_nx_graph_copy(self) -> nx.DiGraph:
        return deepcopy(self._nx_graph)

    def get_nncf_graph_pattern_io(self, match: List[str]) -> NNCFGraphPatternIO:
        """
        Returns an NNCFGraphPatternIO object that describes the input/output nodes and edges of a
        subgraph specified by `match`.

        :param match: A list of node keys specifying a subgraph to be matched. The subgraph to be matched will
        consist of nodes with the same keys that are connected with edges in the order they are listed in the
        `match` list
        :return: NNCFGraphPatternIO object describing the inputs and outputs of the matched subgraph
        """

        in_edge_boundary, out_edge_boundary = NNCFGraph._get_edge_boundaries(match, self._nx_graph)
        boundary = in_edge_boundary + out_edge_boundary
        input_nncf_edges = []
        output_nncf_edges = []

        for nx_edge in boundary:
            from_node_key = nx_edge[0]
            to_node_key = nx_edge[1]
            data = nx_edge[2]
            nncf_edge = NNCFGraphEdge(
                self._nodes[from_node_key],
                self._nodes[to_node_key],
                input_port_id=data[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR],
                output_port_id=data[NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR],
                tensor_shape=data[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR],
                dtype=data[NNCFGraph.DTYPE_EDGE_ATTR],
                parallel_input_port_ids=data[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR],
            )
            if from_node_key in match:
                output_nncf_edges.append(nncf_edge)
            elif to_node_key in match:
                input_nncf_edges.append(nncf_edge)
            else:
                raise nncf.InternalError("Invalid graph expression supplied!")

        return NNCFGraphPatternIO(input_nncf_edges, output_nncf_edges)

    def get_nx_edge(self, node_u: NNCFNode, node_v: NNCFNode) -> OutEdgeView:
        nx_node_u = self._nx_graph.nodes[self._node_id_to_key_dict[node_u.node_id]]
        nx_node_v = self._nx_graph.nodes[self._node_id_to_key_dict[node_v.node_id]]
        return self._nx_graph.edges[nx_node_u["key"], nx_node_v["key"]]

    def get_nodes_count(self) -> int:
        return int(self._nx_graph.number_of_nodes())

    def get_edge(self, from_node: NNCFNode, to_node: NNCFNode) -> NNCFGraphEdge:
        """
        Returns an NNCFGraphEdge object that corresponds to an edge connecting two given NNCFNodes in this
        graph.
        :param from_node: The NNCFNode in this graph that sources the edge.
        :param to_node: The NNCFNode in this graph that sinks the edge.
        :return: The NNCFGraphEdge object representing the edge between `from_node` and `to_node`.
        """
        data = self.get_nx_edge(from_node, to_node)
        return NNCFGraphEdge(
            from_node,
            to_node,
            data[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR],
            data[NNCFGraph.OUTPUT_PORT_ID_EDGE_ATTR],
            data[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR],
            data[NNCFGraph.DTYPE_EDGE_ATTR],
            data[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR],
        )

    def get_all_edges(self) -> Generator[NNCFGraphEdge, None, None]:
        for nx_edge in self._nx_graph.in_edges:
            yield self.get_edge(self.get_node_by_key(nx_edge[0]), self.get_node_by_key(nx_edge[1]))

    def remove_nodes_from(self, nodes: List[NNCFNode]) -> None:
        """
        Removes nodes from the current NNCFGraph instance.
        We use the remove_node method here because remove_nodes_from uses a silent fail instead of an exception.

        :param nodes: List of NNCFNodes to remove.
        """
        for node in nodes:
            self._nx_graph.remove_node(node.node_key)
            del self._nodes[node.node_key]

        self._node_id_to_key_dict = {}
        for node_key, node in self._nx_graph.nodes.items():
            self._node_id_to_key_dict[node["id"]] = node_key  # type:ignore

    def find_matching_subgraphs(self, patterns: GraphPattern, strict: bool = True) -> List[List[NNCFNode]]:
        """
        Returns subgraphs of matched pattern in patterns.

        :param patterns: Instance of GraphPattern containing all patterns.
        :param strict: If True returns only strict matched subgraphs, if False - all matched subgraphs.
        :return: List of subgraphs that are matching by pattern matching.
            Subgraph is a ordered list of nodes of matched subgraph
        The returned nodes order relies on DiGraphMatcher isomorphic subgraphs matching logic from networkX package.
        DiGraphMatcher does not guarantee a specific order for returning isomorphic subgraphs.
        """
        output = []
        for matched_subgraph in find_subgraphs_matching_pattern(self._nx_graph, patterns, strict):
            subgraph_list = []
            for node_key in matched_subgraph:
                subgraph_list.append(self.get_node_by_key(node_key))
            output.append(subgraph_list)
        return output



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from typing import List, Optional, Type

import openvino.runtime as ov

import nncf
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.operator_metatypes import OperatorMetatypeRegistry
from nncf.common.graph.operator_metatypes import UnknownMetatype
from nncf.common.hardware.opset import HWConfigOpName

OV_OPERATOR_METATYPES = OperatorMetatypeRegistry("openvino_operator_metatypes")


class OVOpMetatype(OperatorMetatype):
    op_names: List[str] = []
    subtypes: List[Type[OperatorMetatype]] = []

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return cls.op_names

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return cls.subtypes

    @classmethod
    def matches(cls, node: ov.Node) -> Optional[bool]:
        return node.op_type in cls.op_names

    @classmethod
    def determine_subtype(cls, node: ov.Node) -> Optional[Type[OperatorMetatype]]:
        matches = []
        for subtype in cls.get_subtypes():
            if subtype.matches(node):
                matches.append(subtype)
        if len(matches) > 1:
            raise nncf.InternalError("Multiple subtypes match operator call - can not determine single subtype.")
        if not matches:
            return None
        return matches[0]


@OV_OPERATOR_METATYPES.register()
class OVConvolutionMetatype(OVOpMetatype):
    name = "ConvOp"
    op_names = ["Convolution"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    output_channel_axis = 1


@OV_OPERATOR_METATYPES.register()
class OVConvolutionBackpropDataMetatype(OVOpMetatype):
    name = "ConvBackpropDataOp"
    op_names = ["ConvolutionBackpropData"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    output_channel_axis = 1


@OV_OPERATOR_METATYPES.register(is_subtype=True)
class OVDepthwiseConvolutionMetatype(OVOpMetatype):
    name = "DepthwiseConvolutionOp"
    op_names = ["GroupConvolution"]
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]
    output_channel_axis = 1

    @classmethod
    def matches(cls, node: ov.Node) -> bool:
        return _is_depthwise_conv(node)


@OV_OPERATOR_METATYPES.register()
class OVGroupConvolutionMetatype(OVOpMetatype):
    name = "GroupConvolutionOp"
    op_names = ["GroupConvolution"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    subtypes = [OVDepthwiseConvolutionMetatype]
    output_channel_axis = 1


@OV_OPERATOR_METATYPES.register()
class OVGroupConvolutionBackpropDataMetatype(OVOpMetatype):
    name = "GroupConvolutionBackpropDataOp"
    op_names = ["GroupConvolutionBackpropData"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    output_channel_axis = 1


@OV_OPERATOR_METATYPES.register()
class OVMatMulMetatype(OVOpMetatype):
    name = "MatMulOp"
    op_names = ["MatMul"]
    hw_config_names = [HWConfigOpName.MATMUL]
    output_channel_axis = -1


@OV_OPERATOR_METATYPES.register()
class OVReluMetatype(OVOpMetatype):
    name = "ReluOp"
    op_names = ["Relu"]


@OV_OPERATOR_METATYPES.register()
class OVGeluMetatype(OVOpMetatype):
    name = "GeluOp"
    op_names = ["Gelu"]
    hw_config_names = [HWConfigOpName.GELU]


@OV_OPERATOR_METATYPES.register()
class OVEluMetatype(OVOpMetatype):
    name = "EluOp"
    op_names = ["Elu"]


@OV_OPERATOR_METATYPES.register()
class OVPReluMetatype(OVOpMetatype):
    name = "PReluOp"
    op_names = ["PReLU"]


@OV_OPERATOR_METATYPES.register()
class OVSigmoidMetatype(OVOpMetatype):
    name = "SigmoidOp"
    op_names = ["Sigmoid"]


@OV_OPERATOR_METATYPES.register()
class OVHSigmoidMetatype(OVOpMetatype):
    name = "HSigmoidOp"
    op_names = ["HSigmoid"]


@OV_OPERATOR_METATYPES.register()
class OVHardSigmoidMetatype(OVOpMetatype):
    name = "HardSigmoidOp"
    op_names = ["HardSigmoid"]


@OV_OPERATOR_METATYPES.register()
class OVAvgPoolMetatype(OVOpMetatype):
    name = "AvgPoolOp"
    op_names = ["AvgPool"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@OV_OPERATOR_METATYPES.register()
class OVMaxPoolMetatype(OVOpMetatype):
    name = "MaxPoolOp"
    op_names = ["MaxPool"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@OV_OPERATOR_METATYPES.register()
class OVConstantMetatype(OVOpMetatype):
    name = "ConstantOp"
    op_names = ["Constant"]


@OV_OPERATOR_METATYPES.register()
class OVAddMetatype(OVOpMetatype):
    name = "AddOp"
    op_names = ["Add"]
    hw_config_names = [HWConfigOpName.ADD]


@OV_OPERATOR_METATYPES.register()
class OVSubtractMetatype(OVOpMetatype):
    name = "SubtractOp"
    op_names = ["Subtract"]
    hw_config_names = [HWConfigOpName.SUBTRACT]


@OV_OPERATOR_METATYPES.register()
class OVMultiplyMetatype(OVOpMetatype):
    name = "MultiplyOp"
    op_names = ["Multiply"]
    hw_config_names = [HWConfigOpName.MULTIPLY]


@OV_OPERATOR_METATYPES.register()
class OVDivideMetatype(OVOpMetatype):
    name = "DivideOp"
    op_names = ["Divide"]
    hw_config_names = [HWConfigOpName.DIVIDE]


@OV_OPERATOR_METATYPES.register()
class OVSumMetatype(OVOpMetatype):
    name = "SumOp"
    op_names = ["ReduceSum"]
    hw_config_names = [HWConfigOpName.REDUCESUM]


@OV_OPERATOR_METATYPES.register()
class OVConcatMetatype(OVOpMetatype):
    name = "ConcatOp"
    op_names = ["Concat"]
    hw_config_names = [HWConfigOpName.CONCAT]


@OV_OPERATOR_METATYPES.register()
class OVBatchNormMetatype(OVOpMetatype):
    name = "BatchNormalizationOp"
    op_names = ["BatchNormInference"]


@OV_OPERATOR_METATYPES.register()
class OVInterpolateMetatype(OVOpMetatype):
    name = "InterpolateOp"
    op_names = ["Interpolate"]
    hw_config_names = [HWConfigOpName.INTERPOLATE]


@OV_OPERATOR_METATYPES.register()
class OVMVNMetatype(OVOpMetatype):
    name = "MVNOp"
    op_names = ["MVN"]
    hw_config_names = [HWConfigOpName.MVN]


@OV_OPERATOR_METATYPES.register()
class OVNormalizeL2Metatype(OVOpMetatype):
    name = "NormalizeL2Op"
    op_names = ["NormalizeL2"]


@OV_OPERATOR_METATYPES.register()
class OVReshapeMetatype(OVOpMetatype):
    name = "ReshapeOp"
    op_names = ["Reshape"]
    hw_config_names = [HWConfigOpName.RESHAPE]


@OV_OPERATOR_METATYPES.register()
class OVShapeOfMetatype(OVOpMetatype):
    name = "ShapeOfOp"
    op_names = ["ShapeOf"]


@OV_OPERATOR_METATYPES.register()
class OVNonZeroMetatype(OVOpMetatype):
    name = "NonZeroOp"
    op_names = ["NonZero"]


@OV_OPERATOR_METATYPES.register()
class OVSplitMetatype(OVOpMetatype):
    name = "SplitOp"
    op_names = ["Split"]
    hw_config_names = [HWConfigOpName.SPLIT]


@OV_OPERATOR_METATYPES.register()
class OVVariadicSplitMetatype(OVOpMetatype):
    name = "VariadicSplitOp"
    op_names = ["VariadicSplit"]


@OV_OPERATOR_METATYPES.register()
class OVShuffleChannelsMetatype(OVOpMetatype):
    name = "ShuffleChannelsOp"
    op_names = ["ShuffleChannels"]


@OV_OPERATOR_METATYPES.register()
class OVBroadcastMetatype(OVOpMetatype):
    name = "BroadcastOp"
    op_names = ["Broadcast"]


@OV_OPERATOR_METATYPES.register()
class OVConvertLikeMetatype(OVOpMetatype):
    name = "ConvertLikeOp"
    op_names = ["ConvertLike"]


@OV_OPERATOR_METATYPES.register()
class OVSpaceToBatchMetatype(OVOpMetatype):
    name = "SpaceToBatchOp"
    op_names = ["SpaceToBatch"]


@OV_OPERATOR_METATYPES.register()
class OVBatchToSpaceMetatype(OVOpMetatype):
    name = "BatchToSpaceOp"
    op_names = ["BatchToSpace"]


@OV_OPERATOR_METATYPES.register()
class OVDepthToSpaceMetatype(OVOpMetatype):
    name = "DepthToSpaceOp"
    op_names = ["DepthToSpace"]


@OV_OPERATOR_METATYPES.register()
class OVSpaceToDepthMetatype(OVOpMetatype):
    name = "SpaceToDepthOp"
    op_names = ["SpaceToDepth"]


@OV_OPERATOR_METATYPES.register()
class OVLSTMSequenceMetatype(OVOpMetatype):
    name = "LSTMSequenceOp"
    op_names = ["LSTMSequence"]
    hw_config_names = [HWConfigOpName.LSTMSEQUENCE]
    const_channel_axis = [1]  # const layout: [num_directions, 4 \* hidden_size, input_size]


@OV_OPERATOR_METATYPES.register()
class OVGRUSequenceMetatype(OVOpMetatype):
    name = "GRUSequenceOp"
    op_names = ["GRUSequence"]
    hw_config_names = [HWConfigOpName.GRUSEQUENCE]
    const_channel_axis = [1]  # const layout: [num_directions, 3 \* hidden_size, input_size]


@OV_OPERATOR_METATYPES.register()
class OVFakeQuantizeMetatype(OVOpMetatype):
    name = "FakeQuantizeOp"
    op_names = ["FakeQuantize"]


@OV_OPERATOR_METATYPES.register()
class OVFakeConvertMetatype(OVOpMetatype):
    name = "FakeConvertOp"
    op_names = ["FakeConvert"]


@OV_OPERATOR_METATYPES.register()
class OVLessMetatype(OVOpMetatype):
    name = "LessOp"
    op_names = ["Less"]
    hw_config_names = [HWConfigOpName.LESS]


@OV_OPERATOR_METATYPES.register()
class OVLessEqualMetatype(OVOpMetatype):
    name = "LessEqualOp"
    op_names = ["LessEqual"]
    hw_config_names = [HWConfigOpName.LESSEQUAL]


@OV_OPERATOR_METATYPES.register()
class OVGreaterMetatype(OVOpMetatype):
    name = "GreaterOp"
    op_names = ["Greater"]
    hw_config_names = [HWConfigOpName.GREATER]


@OV_OPERATOR_METATYPES.register()
class OVGreaterEqualMetatype(OVOpMetatype):
    name = "GreaterEqualOp"
    op_names = ["GreaterEqual"]
    hw_config_names = [HWConfigOpName.GREATEREQUAL]


@OV_OPERATOR_METATYPES.register()
class OVEqualMetatype(OVOpMetatype):
    name = "EqualOp"
    op_names = ["Equal"]
    hw_config_names = [HWConfigOpName.EQUAL]


@OV_OPERATOR_METATYPES.register()
class OVNotEqualMetatype(OVOpMetatype):
    name = "NotEqualOp"
    op_names = ["NotEqual"]
    hw_config_names = [HWConfigOpName.NOTEQUAL]


@OV_OPERATOR_METATYPES.register()
class OVLogicalNotMetatype(OVOpMetatype):
    name = "LogicalNotOp"
    op_names = ["LogicalNot"]
    hw_config_names = [HWConfigOpName.LOGICALNOT]


@OV_OPERATOR_METATYPES.register()
class OVLogicalAndMetatype(OVOpMetatype):
    name = "LogicalAndOp"
    op_names = ["LogicalAnd"]
    hw_config_names = [HWConfigOpName.LOGICALAND]


@OV_OPERATOR_METATYPES.register()
class OVLogicalOrMetatype(OVOpMetatype):
    name = "LogicalOrOp"
    op_names = ["LogicalOr"]
    hw_config_names = [HWConfigOpName.LOGICALOR]


@OV_OPERATOR_METATYPES.register()
class OVLogicalXorMetatype(OVOpMetatype):
    name = "LogicalXorOp"
    op_names = ["LogicalXor"]
    hw_config_names = [HWConfigOpName.LOGICALXOR]


@OV_OPERATOR_METATYPES.register(is_subtype=True)
class OVEmbeddingMetatype(OVOpMetatype):
    name = "EmbeddingOp"
    hw_config_names = [HWConfigOpName.EMBEDDING]
    const_channel_axis = [0]

    @classmethod
    def matches(cls, node: ov.Node) -> bool:
        return _is_embedding(node)


@OV_OPERATOR_METATYPES.register()
class OVFloorMetatype(OVOpMetatype):
    name = "FloorOp"
    op_names = ["Floor"]


@OV_OPERATOR_METATYPES.register()
class OVFloorModMetatype(OVOpMetatype):
    name = "FloorModOp"
    op_names = ["FloorMod"]
    hw_config_names = [HWConfigOpName.FLOORMOD]


@OV_OPERATOR_METATYPES.register()
class OVMaximumMetatype(OVOpMetatype):
    name = "MaximumOp"
    op_names = ["Maximum"]
    hw_config_names = [HWConfigOpName.MAXIMUM]


@OV_OPERATOR_METATYPES.register()
class OVMinimumMetatype(OVOpMetatype):
    name = "MinimumOp"
    op_names = ["Minimum"]
    hw_config_names = [HWConfigOpName.MINIMUM]


@OV_OPERATOR_METATYPES.register()
class OVSqrtMetatype(OVOpMetatype):
    name = "SqrtOp"
    op_names = ["Sqrt"]
    hw_config_names = [HWConfigOpName.POWER]


@OV_OPERATOR_METATYPES.register()
class OVPowerMetatype(OVOpMetatype):
    name = "PowerOp"
    op_names = ["Power"]
    hw_config_names = [HWConfigOpName.POWER]


@OV_OPERATOR_METATYPES.register()
class OVLogMetatype(OVOpMetatype):
    name = "LogOp"
    op_names = ["Log"]


@OV_OPERATOR_METATYPES.register()
class OVROIAlignMetatype(OVOpMetatype):
    name = "ROIAlignOp"
    op_names = ["ROIAlign"]


@OV_OPERATOR_METATYPES.register()
class OVROIPoolingMetatype(OVOpMetatype):
    name = "ROIPoolingOp"
    op_names = ["ROIPooling"]


@OV_OPERATOR_METATYPES.register()
class OVGatherMetatype(OVOpMetatype):
    name = "GatherOp"
    op_names = ["Gather"]
    subtypes = [OVEmbeddingMetatype]


@OV_OPERATOR_METATYPES.register()
class OVGatherNDMetatype(OVOpMetatype):
    name = "GatherNDOp"
    op_names = ["GatherND"]


@OV_OPERATOR_METATYPES.register()
class OVGatherElementsMetatype(OVOpMetatype):
    name = "GatherElementsOp"
    op_names = ["GatherElements"]


@OV_OPERATOR_METATYPES.register()
class OVUnsqueezeMetatype(OVOpMetatype):
    name = "UnsqueezeOp"
    op_names = ["Unsqueeze"]
    hw_config_names = [HWConfigOpName.UNSQUEEZE]


@OV_OPERATOR_METATYPES.register()
class OVSqueezeMetatype(OVOpMetatype):
    name = "SqueezeOp"
    op_names = ["Squeeze"]
    hw_config_names = [HWConfigOpName.SQUEEZE]


@OV_OPERATOR_METATYPES.register()
class OVNonMaxSuppressionMetatype(OVOpMetatype):
    name = "NonMaxSuppressionOp"
    op_names = ["NonMaxSuppression"]


@OV_OPERATOR_METATYPES.register()
class OVReduceMinMetatype(OVOpMetatype):
    name = "ReduceMinOp"
    op_names = ["ReduceMin"]


@OV_OPERATOR_METATYPES.register()
class OVReduceMaxMetatype(OVOpMetatype):
    name = "ReduceMaxOp"
    op_names = ["ReduceMax"]
    hw_config_names = [HWConfigOpName.REDUCEMAX]


@OV_OPERATOR_METATYPES.register()
class OVReduceMeanMetatype(OVOpMetatype):
    name = "ReduceMeanOp"
    op_names = ["ReduceMean"]
    hw_config_names = [HWConfigOpName.REDUCEMEAN]


@OV_OPERATOR_METATYPES.register()
class OVReduceL1Metatype(OVOpMetatype):
    name = "ReduceL1Op"
    op_names = ["ReduceL1"]


@OV_OPERATOR_METATYPES.register()
class OVReduceL2Metatype(OVOpMetatype):
    name = "ReduceL2Op"
    op_names = ["ReduceL2"]
    hw_config_names = [HWConfigOpName.REDUCEL2]


@OV_OPERATOR_METATYPES.register()
class OVTopKMetatype(OVOpMetatype):
    name = "TopKOp"
    op_names = ["TopK"]


@OV_OPERATOR_METATYPES.register()
class OVStridedSliceMetatype(OVOpMetatype):
    name = "StridedSliceOp"
    op_names = ["StridedSlice"]
    hw_config_names = [HWConfigOpName.STRIDEDSLICE]


@OV_OPERATOR_METATYPES.register()
class OVExpMetatype(OVOpMetatype):
    name = "ExpOp"
    op_names = ["Exp"]


@OV_OPERATOR_METATYPES.register()
class OVTransposeMetatype(OVOpMetatype):
    name = "TransposeOp"
    op_names = ["Transpose"]
    hw_config_names = [HWConfigOpName.TRANSPOSE]


@OV_OPERATOR_METATYPES.register()
class OVTileMetatype(OVOpMetatype):
    name = "TileOp"
    op_names = ["Tile"]
    hw_config_names = [HWConfigOpName.TILE]


@OV_OPERATOR_METATYPES.register()
class OVScatterElementsUpdateMetatype(OVOpMetatype):
    name = "ScatterElementsUpdateOp"
    op_names = ["ScatterElementsUpdate"]


@OV_OPERATOR_METATYPES.register()
class OVScatterNDUpdateMetatype(OVOpMetatype):
    name = "ScatterNDUpdateOp"
    op_names = ["ScatterNDUpdate"]


@OV_OPERATOR_METATYPES.register()
class OVScatterUpdateMetatype(OVOpMetatype):
    name = "ScatterUpdateOp"
    op_names = ["ScatterUpdate"]


@OV_OPERATOR_METATYPES.register()
class OVSoftmaxMetatype(OVOpMetatype):
    name = "SoftmaxOp"
    op_names = ["SoftMax", "Softmax"]


@OV_OPERATOR_METATYPES.register()
class OVPadMetatype(OVOpMetatype):
    name = "PadOp"
    op_names = ["Pad"]
    hw_config_names = [HWConfigOpName.PAD]


@OV_OPERATOR_METATYPES.register()
class OVReadValueMetatype(OVOpMetatype):
    name = "ReadValueOp"
    op_names = ["ReadValue"]


@OV_OPERATOR_METATYPES.register()
class OVAssignMetatype(OVOpMetatype):
    name = "AssignOp"
    op_names = ["Assign"]


@OV_OPERATOR_METATYPES.register()
class OVConvertMetatype(OVOpMetatype):
    name = "ConvertOp"
    op_names = ["Convert"]


@OV_OPERATOR_METATYPES.register()
@INPUT_NOOP_METATYPES.register()
class OVParameterMetatype(OVOpMetatype):
    name = "ParameterOp"
    op_names = ["Parameter"]


@OV_OPERATOR_METATYPES.register()
@OUTPUT_NOOP_METATYPES.register()
class OVResultMetatype(OVOpMetatype):
    name = "ResultOp"
    op_names = ["Result"]


@OV_OPERATOR_METATYPES.register()
class OVSwishMetatype(OVOpMetatype):
    name = "SwishOp"
    op_names = ["Swish"]


@OV_OPERATOR_METATYPES.register()
class OVHSwishMetatype(OVOpMetatype):
    name = "HSwishhOp"
    op_names = ["HSwish"]


@OV_OPERATOR_METATYPES.register()
class OVClampMetatype(OVOpMetatype):
    name = "ClampOp"
    op_names = ["Clamp"]


@OV_OPERATOR_METATYPES.register()
class OVSquaredDifferenceMetatype(OVOpMetatype):
    name = "SquaredDifferenceOp"
    op_names = ["SquaredDifference"]


@OV_OPERATOR_METATYPES.register()
class OVDeformableConvolutionMetatype(OVOpMetatype):
    name = "DeformableConvolutionOp"
    op_names = ["DeformableConvolution"]


@OV_OPERATOR_METATYPES.register()
class OVAbsMetatype(OVOpMetatype):
    name = "AbsOp"
    op_names = ["Abs"]


@OV_OPERATOR_METATYPES.register()
class OVIfMetatype(OVOpMetatype):
    name = "IfOp"
    op_names = ["If"]


@OV_OPERATOR_METATYPES.register()
class OVGroupNormalizationMetatype(OVOpMetatype):
    name = "GroupNormalizationOp"
    op_names = ["GroupNormalization"]
    hw_config_names = [HWConfigOpName.GROUPNORMALIZATION]


@OV_OPERATOR_METATYPES.register()
class OVScaledDotProductAttentionMetatype(OVOpMetatype):
    name = "ScaledDotProductAttentionOp"
    op_names = ["ScaledDotProductAttention"]
    hw_config_names = [HWConfigOpName.SCALED_DOT_PRODUCT_ATTENTION]
    target_input_ports = [0, 1]


def get_operator_metatypes() -> List[Type[OperatorMetatype]]:
    """
    Returns a list of the operator metatypes.
    :return: List of operator metatypes .
    """
    return list(OV_OPERATOR_METATYPES.registry_dict.values())


def get_operation_const_op(operation: ov.Node, const_port_id: int) -> Optional[ov.Node]:
    """
    Returns constant node of given operation placed on given const port id.

    :param operation: Given operation.
    :param const_port_id: Given constant port id.
    :returns: Constant node of given operation placed on given const port id.
    """
    node = operation.input_value(const_port_id).get_node()

    # There are several cases here
    # (Constant) -> (Operation)
    # (Constant) -> (Convert) -> (Operation)
    # (Constant) -> (Convert) -> (FakeQuantize, FakeConvert) -> (Operation)
    # (Constant) -> (Convert) -> (FakeQuantize, FakeConvert) -> (Reshape) -> (Operation)
    #  and etc. We need properly find the constant node. So we start with
    # `node` and traverse up until the constant node is not found.
    queue = deque([node])
    constant_node = None
    allowed_propagation_types_list = ["Convert", "FakeQuantize", "FakeConvert", "Reshape"]

    while len(queue) != 0:
        curr_node = queue.popleft()
        if curr_node.get_type_name() == "Constant":
            constant_node = curr_node
            break
        if len(curr_node.inputs()) == 0:
            break
        if curr_node.get_type_name() in allowed_propagation_types_list:
            queue.append(curr_node.input_value(0).get_node())

    return constant_node


def _is_depthwise_conv(node: ov.Node) -> bool:
    """
    Returns True if the group convolution is depthwise, False - otherwise.
    Depthwise convolution is a convolution satisfies the following rule:
    groups == in_channels and inp_channels > 1.
    Weight tensor layout is [groups, output channels / groups, input channels / groups, Z, Y, X],
    where Z, Y, X - spatial axes.

    :param node: GroupConvolution node to check whether it is depthwise.
    :return: True if the convolution is depthwise, False - otherwise.
    """
    inp_channels = node.input_value(0).get_partial_shape().get_dimension(1)
    groups = node.input_value(1).get_partial_shape().get_dimension(0)
    if inp_channels.is_dynamic or groups.is_dynamic:
        return False
    inp_channels = inp_channels.get_length()
    groups = groups.get_length()
    return groups == inp_channels and inp_channels > 1


def _is_embedding(node: ov.Node) -> bool:
    """
    Returns True if the layer can be represented as embedding, False - otherwise.

    :param node: Layer to check whether it is embedding.
    :return: True if the layer is embedding, False - otherwise.
    """
    allowed_types_list = ["f16", "f32", "f64"]
    const_port_id = 0
    input_tensor = node.input_value(const_port_id)
    if input_tensor.get_element_type().get_type_name() in allowed_types_list:
        const_node = get_operation_const_op(node, const_port_id)
        if const_node is not None:
            return True

    return False


def get_node_metatype(node: ov.Node) -> Type[OperatorMetatype]:
    """
    Determine NNCF meta type for OpenVINO node.

    :param node: OpenVINO node.
    :return: NNCF meta type which corresponds to OpenVINO node.
    """
    node_type = node.get_type_name()
    metatype = OV_OPERATOR_METATYPES.get_operator_metatype_by_op_name(node_type)
    if metatype is not UnknownMetatype and metatype.get_subtypes():
        subtype = metatype.determine_subtype(node)
        if subtype is not None:
            metatype = subtype
    return metatype



# Copyright (c) 2023 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import torch

import nncf.torch.graph.operator_metatypes as om
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.quantization.quantizer_propagation.structs import QuantizationTrait
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.experimental.common.tensor_statistics.collectors import MaxAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.tensor import Tensor
from nncf.openvino.graph.transformations.commands import OVMultiplyInsertionCommand
from nncf.openvino.graph.transformations.commands import OVWeightUpdateCommand
from nncf.quantization.algorithms.smooth_quant.backend import SmoothQuantAlgoBackend
from nncf.torch.graph.transformations.command_creation import create_command_to_update_weight
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.layer_utils import COMPRESSION_MODULES
from nncf.torch.layer_utils import CompressionParameter
from nncf.torch.layer_utils import StatefullModuleInterface
from nncf.torch.model_graph_manager import get_const_data
from nncf.torch.model_graph_manager import get_const_node
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.default_quantization import DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT
from nncf.torch.tensor_statistics.collectors import PTAbsMaxReducer
from nncf.torch.tensor_statistics.collectors import PTNNCFCollectorTensorProcessor


@COMPRESSION_MODULES.register()
class SQMultiply(torch.nn.Module, StatefullModuleInterface):
    SCALE_SHAPE_KEY = "scale_shape"

    def __init__(self, scale_shape: Tuple[int, ...]):
        super().__init__()
        self._scale_value = CompressionParameter(torch.empty(scale_shape))

    @property
    def scale(self) -> torch.nn.Parameter:
        return self._scale_value

    @scale.setter
    def scale(self, value: torch.tensor):
        self._scale_value.data = value

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.mul(x, self._scale_value)

    def get_config(self) -> Dict[str, Any]:
        return {self.SCALE_SHAPE_KEY: list(self._scale_value.shape)}

    @classmethod
    def from_config(cls, state) -> "SQMultiply":
        return SQMultiply(state[cls.SCALE_SHAPE_KEY])


PT_PRE_LAYER_TARGET_TYPE = TargetType.OPERATOR_PRE_HOOK


class PTSmoothQuantAlgoBackend(SmoothQuantAlgoBackend):
    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.PTConv1dMetatype,
            om.PTConv2dMetatype,
            om.PTConv3dMetatype,
        ]

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return [om.PTLinearMetatype]

    @property
    def quantize_agnostic_metatypes(self) -> List[OperatorMetatype]:
        return DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT[QuantizationTrait.QUANTIZATION_AGNOSTIC]

    @staticmethod
    def pre_layer_target_type() -> TargetType:
        return PT_PRE_LAYER_TARGET_TYPE

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        return PTTargetPoint(target_type, target_node_name, input_port_id=port_id)

    @staticmethod
    def is_node_with_weights(node: NNCFNode) -> bool:
        # Metatypes of linears and convolutions guarantee
        # all nodes with the metatypes have weights, we can skip
        # this check by returning True.
        return True

    @staticmethod
    def get_activations_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        # Metatypes of linears and convolutions guarantee
        # all nodes with the metatypes have 0 activation port id.
        return 0

    @staticmethod
    def get_abs_max_channel_collector(
        num_samples: int, stats_reduction_axes: Tuple[int], inplace: bool, branch_key: str
    ) -> TensorCollector:
        collector = TensorCollector()
        reducer = PTAbsMaxReducer(reduction_axes=stats_reduction_axes)
        aggregator = MaxAggregator(tensor_processor=PTNNCFCollectorTensorProcessor, num_samples=num_samples)
        collector.register_statistic_branch(branch_key, reducer, aggregator)
        return collector

    @staticmethod
    def get_weight_value(node_with_weight: NNCFNode, model: NNCFNetwork) -> Tensor:
        weight_node = get_const_node(
            node_with_weight, node_with_weight.metatype.weight_port_ids[0], model.nncf.get_graph()
        )
        if weight_node is None:
            raise RuntimeError(f"{node_with_weight} node has no weight node.")
        weight_data = get_const_data(weight_node, model)
        return Tensor(weight_data)

    @staticmethod
    def get_weight_tensor_port_id(node: NNCFNode) -> int:
        const_ids = node.layer_attributes.get_const_port_ids()
        if len(const_ids) != 1:
            raise RuntimeError(f"Found more than 1 port for {node.node_name} node")
        return const_ids[0]

    @staticmethod
    def weight_update_command(node_with_weight: NNCFNode, weight_value: np.ndarray) -> OVWeightUpdateCommand:
        return create_command_to_update_weight(node_with_weight, weight_value)

    @staticmethod
    def scale_insertion_command(
        source_node: NNCFNode,
        scale_value: torch.Tensor,
        source_output_port_id: int,
        nodes: List[NNCFNode],
        scale_node_name: str,
    ) -> OVMultiplyInsertionCommand:
        input_port_id = 0
        target_points = []
        for node in nodes:
            target_points.append(PTTargetPoint(PT_PRE_LAYER_TARGET_TYPE, node.node_name, input_port_id=input_port_id))

        sq_multiply = SQMultiply(scale_value.shape)
        sq_multiply.scale = scale_value
        return PTSharedFnInsertionCommand(target_points, sq_multiply, scale_node_name)

    @staticmethod
    def get_activation_channel_axis(node: NNCFNode, port_id: int) -> int:
        if node.metatype == om.PTLinearMetatype:
            return -1
        # TODO: Add activation axis calculation when MatMul will be supported
        return 1

    @staticmethod
    def get_weight_channel_axis(node: NNCFNode) -> int:
        # TODO: Add activation axis calculation when MatMul will be supported
        return 1

    @staticmethod
    def is_node_with_shared_weight(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return node.is_shared()

    @staticmethod
    def get_filter_fn_for_statistics(activation_port_id: int, algorithm_key: str) -> Callable[[StatisticPoint], bool]:
        def filter_func(point: StatisticPoint) -> bool:
            return (
                algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == PT_PRE_LAYER_TARGET_TYPE
                and point.target_point.input_port_id == activation_port_id
            )

        return filter_func



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from nncf.common.tensor_statistics.statistics import MeanTensorStatistic
from nncf.common.tensor_statistics.statistics import MinMaxTensorStatistic
from nncf.common.tensor_statistics.statistics import RawTensorStatistic


class OVMinMaxTensorStatistic(MinMaxTensorStatistic):
    @staticmethod
    def tensor_eq(tensor1: np.ndarray, tensor2: np.ndarray, rtol=1e-6) -> bool:
        return bool(np.allclose(tensor1, tensor2, rtol=rtol))


class OVMeanTensorStatistic(MeanTensorStatistic):
    @staticmethod
    def tensor_eq(tensor: np.ndarray, rtol=1e-6) -> bool:
        return bool(np.all(tensor, rtol=rtol))


class OVRawTensorStatistic(RawTensorStatistic):
    @staticmethod
    def tensor_eq(tensor: np.ndarray, rtol=1e-6) -> bool:
        return bool(np.all(tensor, rtol=rtol))



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List

import networkx as nx  # type:ignore
import networkx.algorithms.isomorphism as ism  # type:ignore

from nncf.common.graph.patterns import GraphPattern

ATTRS_TO_SKIP = [GraphPattern.LABEL_ATTR, GraphPattern.PATTERN_NODE_TO_EXCLUDE]


def _are_nodes_matched(node_1, node_2) -> bool:  # type:ignore
    for attr in node_2:
        if attr in ATTRS_TO_SKIP:
            continue
        if attr == GraphPattern.METATYPE_ATTR:
            # GraphPattern.ANY_PATTERN_NODE_TYPE and GraphPattern.NON_PATTERN_NODE_TYPE
            # are matched to any node type.
            if GraphPattern.ANY_PATTERN_NODE_TYPE in node_2[attr] or GraphPattern.NON_PATTERN_NODE_TYPE in node_2[attr]:
                continue
            # Torch and TF pattern mapping based on 'type' section,
            # While ONNX mapping based on metatypes -
            # to support all of them, we need to check the existence of the attributes
            if GraphPattern.NODE_TYPE_ATTR in node_1 and node_1[GraphPattern.NODE_TYPE_ATTR] in node_2[attr]:
                continue
        if node_1[attr] not in node_2[attr]:
            return False
    return True


def _sort_patterns_by_len(pattern: nx.DiGraph) -> int:
    """
    Sort patterns by their length. GraphPattern.NON_PATTERN_NODE_TYPE is not counted as a pattern node.
    """
    non_pattern_nodes = [
        node_id
        for node_id, node_data in pattern.nodes(data=True)
        if GraphPattern.NON_PATTERN_NODE_TYPE in node_data.get(GraphPattern.METATYPE_ATTR, [])
    ]
    return len(pattern) - len(non_pattern_nodes)


def _is_subgraph_matching_strict(graph: nx.DiGraph, pattern: nx.DiGraph, subgraph: Dict[str, str]) -> bool:
    """
    Checks out whether the matched subgraph has:
    1) External predecessors of starting nodes.
    2) External successors of the last nodes.
    3) External successors or predecessors of the nodes which are not starting and last.
    If any of these conditions is True, than returns False, otherwise - True.
    The checks are skipped for NON_PATTERN_NODE_TYPE.
    Example:
    This subgraph matching is not strict.
    (conv2d + BN + ReLU pattern):
            ...
             |
          (conv2d)
             |------\
            (BN)    |
             |      |
           (RELU)   |
             |      |
           (cat)----/
             |
            ...
    :param graph: The model graph.
    :param pattern: The matched pattern.
    :param subgraph: A subgraph of the model graph including the nodes outside the pattern.
    :return: If any of three conditions is True than returns False, otherwise - True.
    """
    starting_nodes = []
    last_nodes = []
    for node in pattern.nodes:
        if not pattern.pred[node] and pattern.succ[node]:
            starting_nodes.append(node)
        if pattern.pred[node] and not pattern.succ[node]:
            last_nodes.append(node)

    for node_from_graph, node_from_pattern in subgraph.items():
        if GraphPattern.NON_PATTERN_NODE_TYPE in pattern.nodes[node_from_pattern].get(GraphPattern.METATYPE_ATTR, []):
            continue
        predecessors_keys = graph.pred[node_from_graph].keys()
        successor_keys = graph.succ[node_from_graph].keys()
        has_external_successors = any(successor_key not in subgraph for successor_key in successor_keys)
        has_external_predcessors = any(predecessor_key not in subgraph for predecessor_key in predecessors_keys)
        if node_from_pattern in starting_nodes and has_external_successors:
            return False
        if node_from_pattern in last_nodes and has_external_predcessors:
            return False
        if (node_from_pattern not in last_nodes and node_from_pattern not in starting_nodes) and (
            has_external_successors or has_external_predcessors
        ):
            return False
    return True


def _copy_subgraph_excluding_non_pattern_node(subgraph: Dict[str, str], pattern_graph: GraphPattern) -> Dict[str, str]:
    """
    Copies a matching subgraph excluding the nodes having GraphPattern.NON_PATTERN_NODE_TYPE
       or GraphPattern.PATTERN_NODE_TO_EXCLUDE.

    :param subgraph: Subgraph
    :param pattern_graph: A graph consists of patterns to match.
    :return: New subgraph without excluded nodes.
    """
    output = {}
    for node_from_graph, node_from_pattern in subgraph.items():
        pattern_node = pattern_graph.graph.nodes[node_from_pattern]
        pattern_node_types = pattern_node.get(GraphPattern.METATYPE_ATTR, [])
        if GraphPattern.NON_PATTERN_NODE_TYPE in pattern_node_types:
            continue
        if pattern_node.get(GraphPattern.PATTERN_NODE_TO_EXCLUDE, False):
            continue
        output[node_from_graph] = node_from_pattern

    return output


def find_subgraphs_matching_pattern(
    graph: nx.DiGraph, pattern_graph: GraphPattern, strict: bool = True
) -> List[List[str]]:
    """
    Finds a list of nodes which define a subgraph matched a pattern in pattern_graph.
    Nodes in each subgraph is stored in lexicographical_topological_sort.

    :param graph: The model graph.
    :param pattern_graph: A graph consists of patterns to match.
    :param strict: If True returns only strict matched subgraphs, if False - all matched subgraphs.
    :return: A list of subgraphs are matched to the patterns. Each subgraph is defined as a list of node keys.
    """
    subgraphs = []
    matched_nodes = set()
    patterns = pattern_graph.get_weakly_connected_subgraphs()
    patterns = sorted(patterns, key=_sort_patterns_by_len, reverse=True)
    for pattern in patterns:
        matcher = ism.DiGraphMatcher(graph, pattern, node_match=_are_nodes_matched)
        for subgraph in matcher.subgraph_isomorphisms_iter():
            if strict and not _is_subgraph_matching_strict(graph, pattern, subgraph):
                continue

            subgraph = _copy_subgraph_excluding_non_pattern_node(subgraph, pattern_graph)
            is_any_node_matched = any(node in matched_nodes for node in subgraph)

            if is_any_node_matched:
                continue

            matched_nodes.update(subgraph)
            sorted_nodes_subgraph = list(nx.lexicographical_topological_sort(graph.subgraph(subgraph)))
            subgraphs.append(sorted_nodes_subgraph)

    return subgraphs



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, List, Optional, OrderedDict, Tuple, TypeVar

import nncf
from nncf import Dataset
from nncf.common.factory import StatisticsAggregatorFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging import nncf_logger
from nncf.common.logging.track_progress import track
from nncf.common.scopes import should_consider_scope
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.common.utils.helpers import create_table
from nncf.experimental.tensor import Tensor
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.parameters import CompressWeightsMode
from nncf.parameters import SensitivityMetric
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.weight_compression.awq import AWQ
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.gptq import GPTQ
from nncf.quantization.algorithms.weight_compression.mixed_precision import MIXED_PRECISION_CRITERIA
from nncf.quantization.algorithms.weight_compression.scale_estimation import ScaleEstimation
from nncf.quantization.algorithms.weight_compression.weight_lowering import WeightCompressionConfig
from nncf.scopes import IgnoredScope
from nncf.scopes import get_ignored_node_names_from_ignored_scope

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")


class WeightCompression(Algorithm):
    """
    Post-training Weight Compression algorithm implementation.

    Compresses weights of Linear and Embedding layers to 8-bit integer or
    to nf4 depending on mode, ratio and group size.
    """

    def __init__(
        self,
        mode: CompressWeightsMode,
        ratio: float,
        group_size: int,
        ignored_scope: IgnoredScope,
        all_layers: bool,
        sensitivity_metric: SensitivityMetric,
        awq: bool,
        subset_size: int,
        scale_estimation: bool,
        gptq: bool,
        advanced_parameters: Optional[AdvancedCompressionParameters] = None,
    ):
        """
        :param mode: Defines a mode for weight compression.
            INT8_SYM stands for 8-bit integer symmetric quantization of all weights.
                Weights are quantized symmetrically with a fixed zero point equals to 128.
            INT8_ASYM is the same as INT8_SYM mode, but weights are quantized to a primary precision asymmetrically
                with a typical non-fixed zero point.
            INT4_SYM stands for a mixed-precision weights quantization with 4-bit integer as a primary precision.
                Weights are quantized to a primary precision symmetrically with a fixed zero point equals to 8.
                All embeddings and the last layer are always compressed to a backup precision, which is INT8_ASYM,
                by default. All others are quantized whether to 4-bit integer or to a backup precision depending on
                criteria and the given ratio.
            INT4_ASYM is the same as INT4_SYM mode, but weights are quantized to a primary precision asymmetrically
                with a typical non-fixed zero point.
            NF4 is the same as INT4_SYM mode, but primary precision is NF4 data type without zero point.
        :param ratio: the ratio between primary and backup precisions (e.g. 0.9 means 90% of layers quantized to NF4
            and the rest to INT8_ASYM).
        :param group_size: number of weights (e.g. 128) in the channel dimension
            that share quantization parameters (scale). The value -1 means no grouping.
        :param ignored_scope: An ignored scope that defined the list of model control
            flow graph nodes to be ignored during quantization.
        :param all_layers: Indicates whether embeddings and last MatMul layers should be compressed to a primary
            precision. By default, the backup precision is assigned for the embeddings and last MatMul layers.
        :param sensitivity_metric: The sensitivity metric for assigning quantization precision to layers. In order to
            preserve the accuracy of the model, the more sensitive layers receives a higher precision.
        :param awq: determines whether to use or not modified AWQ algorithm.
        :param subset_size: Number of data samples to calculate activation statistics used for assigning different
            quantization precision.
        :param scale_estimation: determines whether to use or not scale estimation for 4 bit layers.
        :param gptq: determines whether to use or not GPTQ algorithm.
        :param advanced_parameters: advanced parameters for algorithms in compression pipeline.
        """
        super().__init__()
        self._mode = mode
        self._group_size = group_size
        self._ratio = ratio
        self._ignored_scope = ignored_scope
        self._backend_entity = None
        self._algorithm_key = f"CW_{hash(self)}"
        self._fp_inputs = defaultdict(list)
        self._all_layers = all_layers
        self._sensitivity_metric = sensitivity_metric
        self._awq = awq
        self._subset_size = subset_size
        self._scale_estimation = scale_estimation
        self._gptq = gptq
        self._advanced_parameters = (
            advanced_parameters if advanced_parameters is not None else AdvancedCompressionParameters()
        )

        if self._gptq:
            gptq_params = self._advanced_parameters.gptq_params
            self._gptq_algo = GPTQ(gptq_params.damp_percent, gptq_params.block_size, gptq_params.subset_size)
            self._gptq_statistics = None

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO, BackendType.TORCH]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVWeightCompressionAlgoBackend

            self._backend_entity = OVWeightCompressionAlgoBackend(model)
        elif model_backend == BackendType.TORCH:
            from nncf.quantization.algorithms.weight_compression.torch_backend import PTWeightCompressionAlgoBackend

            self._backend_entity = PTWeightCompressionAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def _get_nodes_to_compress(self, nncf_graph: NNCFGraph) -> List[NNCFNode]:
        """
        Collects nodes in the model's graph corresponding to the layers for weight compression.

        :param nncf_graph: NNCFGraph instance.
        :return: List with the data for each layer.
        """
        weighted_metatypes = (
            self._backend_entity.matmul_metatypes
            + self._backend_entity.embedding_metatypes
            + self._backend_entity.convolution_metatypes
        )

        ordered_nodes_to_compress = []
        ignored_names = get_ignored_node_names_from_ignored_scope(
            self._ignored_scope, nncf_graph, strict=self._ignored_scope.validate
        )
        for node in nncf_graph.topological_sort():
            is_node_with_weights = self._backend_entity.is_node_with_weights(node, nncf_graph)
            is_within_scope = should_consider_scope(node.node_name, ignored_names)
            if node.metatype in weighted_metatypes and is_node_with_weights and is_within_scope:
                ordered_nodes_to_compress.append(node)
        return ordered_nodes_to_compress

    def _get_ratio_defining_params(
        self, all_weight_params: List[WeightCompressionParameters], is_last_layer_shared: bool
    ) -> List[WeightCompressionParameters]:
        """
        Returns the information about weights that are used for ratio calculation between primary
        and backup precisions.

        :param all_weight_params: List of all weight parameters.
        :param is_last_layer_shared: Indicates whether the last layer which shares the weight
            should be quantized or not.
        :return: Information about each weight node that is considered for mixed precision.
        """
        if self._mode in [CompressWeightsMode.INT8_SYM, CompressWeightsMode.INT8_ASYM]:
            return all_weight_params

        ratio_defining_params = list(
            filter(
                lambda wp: wp.node_with_weight.metatype in self._backend_entity.matmul_metatypes,
                all_weight_params,
            )
        )

        # The last MatMul layer is quantized to 4-bits if all_layers=True
        if not self._all_layers and not is_last_layer_shared:
            ratio_defining_params = ratio_defining_params[:-1]

        # Embedding layers are quantized to 4-bits only if all_layers=True.
        if self._all_layers:
            embedding_params = list(
                filter(
                    lambda wp: wp.node_with_weight.metatype in self._backend_entity.embedding_metatypes
                    and len(wp.reduction_axes) == 1,
                    all_weight_params,
                )
            )
            ratio_defining_params.extend(embedding_params)

        return ratio_defining_params

    def _set_weight_compression_config(
        self,
        ratio_defining_params: List[WeightCompressionParameters],
        model: TModel,
        graph: NNCFGraph,
        activations: Optional[Dict[str, List[Tensor]]] = None,
    ) -> None:
        """
        Sets the appropriate compression configuration for weights based on some criteria.

        :param ratio_defining_params: Information about weights that are used for calculating ratio between primary and
            backup precisions.
        :param model: The model.
        :param graph: The model graph associated with the model.
        :param activations: The input activations of the layers considered for compression.
        """
        primary_config = WeightCompressionConfig(mode=self._mode, group_size=self._group_size)
        if self._ratio == 1:
            for weight_param in ratio_defining_params:
                weight_param.compression_config = primary_config
        else:
            criterion_cls = MIXED_PRECISION_CRITERIA.get(self._sensitivity_metric)
            criterion = criterion_cls(
                model, graph, self._backend_entity, ratio_defining_params, primary_config, self._ratio, activations
            )
            criterion.assign_mixed_precision()

    @staticmethod
    def _proportion_str(num_weights_list: List[int], total_num_weights: int, total_num_params: int) -> str:
        """
        Generates a string with proportion between target parameters and all model parameters by number of weights.

        :param num_weights_list: List of number of weights of target model parameters.
        :param total_num_weights: The total number of weights.
        :param total_num_params: The total number of model parameters.
        :return: The string with proportion between target parameters and all model parameters by number of weights.
        """
        percentage = sum(num_weights_list) / max(total_num_weights, 1) * 100
        return f"{percentage:.0f}% ({len(num_weights_list)} / {total_num_params})"

    def _get_bitwidth_distribution_str(
        self, all_params: List[WeightCompressionParameters], ratio_defining_params: List[WeightCompressionParameters]
    ) -> str:
        """
        Generates a table that shows the ratio of weights quantized to different number of bits.

        :param all_params: Information about each weight node.
        :param ratio_defining_params: Information about weights that are used for calculating ratio between primary and
            backup precisions.
        :return: A string containing the table.
        """
        num_bits_vs_num_weights_map = {}
        ratio_defining_weight_names = set(wp.weight_name for wp in ratio_defining_params)
        for data in all_params:
            num_bits = data.compression_config.num_bits
            n_total, n_ratio_defining = num_bits_vs_num_weights_map.get(num_bits, ([], []))
            if data.weight_name in ratio_defining_weight_names:
                n_ratio_defining.append(data.num_weights)
            n_total.append(data.num_weights)
            num_bits_vs_num_weights_map[num_bits] = (n_total, n_ratio_defining)

        num_ratio_defining_weights = sum(ws.num_weights for ws in ratio_defining_params)
        num_ratio_defining_params = len(ratio_defining_params)
        num_total_weights = sum(ws.num_weights for ws in all_params)
        num_params = len(all_params)
        num_bits_vs_num_weights_map = OrderedDict(sorted(num_bits_vs_num_weights_map.items(), reverse=True))
        # Table creation
        header = ["Num bits (N)", "% all parameters (layers)", "% ratio-defining parameters (layers)"]
        rows = []
        for bitwidth, (n_total, n_ratio_defining) in num_bits_vs_num_weights_map.items():
            rows.append(
                [
                    bitwidth,
                    self._proportion_str(n_total, num_total_weights, num_params),
                    self._proportion_str(n_ratio_defining, num_ratio_defining_weights, num_ratio_defining_params),
                ]
            )

        table = create_table(header, rows)
        pretty_string = f"Statistics of the bitwidth distribution:\n{table}"
        return pretty_string

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        self._set_backend_entity(model)
        nodes_to_compress = self._get_nodes_to_compress(graph)

        activations = {}
        if dataset is not None and self._sensitivity_metric != SensitivityMetric.WEIGHT_QUANTIZATION_ERROR:
            activations = self._get_activations(dataset, self._subset_size, nodes_to_compress, graph, model)
        all_weight_params: List[WeightCompressionParameters] = []
        weight_names = set()

        is_last_layer_shared = False
        n = len(nodes_to_compress)
        for i, node in enumerate(nodes_to_compress):
            for weight_name, weight_port_id in self._backend_entity.get_weight_names_and_port_ids(node, graph):
                if weight_name in weight_names:
                    if i == n - 1:
                        is_last_layer_shared = True
                    continue

                weight = self._backend_entity.get_weight(node, weight_port_id, model, graph)
                if weight.dtype not in [
                    TensorDataType.float16,
                    TensorDataType.bfloat16,
                    TensorDataType.float32,
                    TensorDataType.float64,
                ]:
                    continue
                reduction_axes = self._backend_entity.get_reduction_axes(node, weight_port_id, graph)
                if (
                    self._group_size != -1
                    and self._all_layers
                    and node.metatype in self._backend_entity.embedding_metatypes
                    and isinstance(reduction_axes, tuple)
                    and len(reduction_axes) != 1
                ):
                    # NNCF supports multiple reduction axes only for ops with group_size != -1.
                    # Convolution ops are always quantized to 8-bits (without groups).
                    # Embedding layers are quantized to 4-bits only if all_layers=True.
                    # MatMul ops can't have multiple reduction axes.
                    nncf_logger.warning(
                        f"Weight compression expects a single reduction axis, but {len(reduction_axes)} given. "
                        f"Weight shape: {weight.shape}, reduction axes: {reduction_axes}, "
                        f"node name: {node.node_name}. The node will be asymmetrically quantized to 8 bits."
                    )

                weight_params = WeightCompressionParameters(
                    weight_name, node, weight_port_id, weight.size, reduction_axes
                )
                all_weight_params.append(weight_params)
                weight_names.add(weight_name)

        ratio_defining_params = self._get_ratio_defining_params(all_weight_params, is_last_layer_shared)
        self._set_weight_compression_config(ratio_defining_params, model, graph, activations)
        nncf_logger.info(self._get_bitwidth_distribution_str(all_weight_params, ratio_defining_params))

        if self._awq and activations is not None and self._mode != CompressWeightsMode.NF4:
            awq_params = self._advanced_parameters.awq_params
            awq_algo = AWQ(
                model,
                self._backend_entity.name_to_node_mapping,
                all_weight_params,
                nodes_to_compress,
                activations,
                awq_params.subset_size,
                awq_params.percent_to_apply,
                awq_params.alpha_min,
                awq_params.alpha_max,
                awq_params.steps,
            )
            awq_algo.apply(model, graph)

        scales = {}
        zero_points = {}
        if self._scale_estimation and activations is not None and self._mode != CompressWeightsMode.NF4:
            scale_estimation_params = self._advanced_parameters.scale_estimation_params
            scale_algo = ScaleEstimation(
                model,
                self._backend_entity.name_to_node_mapping,
                all_weight_params,
                nodes_to_compress,
                activations,
                scale_estimation_params.subset_size,
                scale_estimation_params.initial_steps,
                scale_estimation_params.scale_steps,
                scale_estimation_params.weight_penalty,
            )
            scales = scale_algo.apply(model, graph)

        if self._gptq:
            model, scales, zero_points = self._gptq_algo.apply(
                model=model,
                graph=graph,
                dataset=dataset,
                weight_compression_parameters=all_weight_params,
                statistic_points=self._gptq_statistics,
                backend_entity=self._backend_entity,
            )

        # Compress model using weight compression parameters
        transformed_model = self._backend_entity.transform_model(
            model,
            graph,
            track(all_weight_params, description="Applying Weight Compression"),
            scales,
            zero_points,
        )

        self._backend_entity.dump_parameters(
            model,
            parameters={
                "mode": self._mode.value,
                "group_size": self._group_size,
                "ratio": self._ratio,
                "all_layers": self._all_layers,
                "ignored_scope": self._ignored_scope,
                "sensitivity_metric": self._sensitivity_metric.value,
                "awq": self._awq,
                "scale_estimation": self._scale_estimation,
                "gptq": self._gptq,
            },
            algo_name="weight_compression",
        )
        return transformed_model

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        pass

    def _get_activation_node_and_port(self, node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[NNCFNode, int]:
        """
        This method returns the activation layer and corresponding port id for the node.

        :param node: NNCFGraph node for which the activation is sought.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Tuple with the activation node and port id.
        """
        activation_port = self._backend_entity.get_activation_port_id(node, nncf_graph)
        activation_edge = nncf_graph.get_input_edges(node)[activation_port]
        activation_node = activation_edge.from_node
        port_id = activation_edge.output_port_id
        return activation_node, port_id

    def _get_fp_inputs(self, statistic_points: StatisticPointsContainer, node_name: str, port_id: int) -> List[Tensor]:
        """
        Collects floating-point statistics for the given node and port id.

        :param statistic_points: Filled StatisticPointsContainer.
        :param node_name: Name of the current layer.
        :param port_id: Port id for statistics collection.
        :return: Collected list of tensor data.
        """

        def input_filter_func(point):
            # For the floating-point statistics collected in POST_LAYER style,
            # we also need to determine the output port id.
            # For the cases when the layer has more than one (0) output port.
            return (
                self._algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == TargetType.POST_LAYER_OPERATION
                and point.target_point.port_id == port_id
            )

        input_id = (node_name, port_id)
        if input_id in self._fp_inputs:
            return self._fp_inputs[input_id]

        input_fp = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name, input_filter_func, self._algorithm_key
        ):
            for value in tensor_collector.get_statistics().values:
                input_fp.append(Tensor(value))
        self._fp_inputs[input_id] = input_fp
        return self._fp_inputs[input_id]

    def _get_activations(
        self, dataset: Dataset, subset_size: int, nodes_to_compress: List[NNCFNode], graph: NNCFGraph, model: TModel
    ) -> Dict[str, List[Tensor]]:
        """
        Collects input activations for the given nodes on the dataset.

        :param dataset: Dataset to collect values.
        :param subset_size: Number of data samples to calculate activation statistics used for assigning different
            quantization precision.
        :param nodes_to_compress: List of nodes, whose inputs are collected.
        :param model: Model for statistics collection.
        :param graph: Model graph.
        :return: statistics values itself per node name.
        """
        activations = {}
        _collected_stat_inputs_map = {}
        statistic_container = StatisticPointsContainer()
        all_act_nodes = set()
        act_vs_shared_node_names_mapping = defaultdict(list)
        matmul_metatypes = self._backend_entity.matmul_metatypes
        filtered_nodes = filter(lambda node: node.metatype in matmul_metatypes, nodes_to_compress)
        for node in filtered_nodes:
            act_node, output_port_id = self._get_activation_node_and_port(node, graph)
            act_node_name = act_node.node_name
            if act_node_name in all_act_nodes:
                act_vs_shared_node_names_mapping[act_node_name].append(node.node_name)
                continue
            all_act_nodes.add(act_node_name)
            output_id = (act_node_name, output_port_id)
            _collected_stat_inputs_map[node.node_name] = output_id

            statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, act_node_name, port_id=output_port_id
            )
            stat_collector = self._backend_entity.raw_statistic_collector(num_samples=subset_size)
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                )
            )

        statistics_aggregator = StatisticsAggregatorFactory.create(model, dataset)
        statistics_aggregator.register_statistic_points(statistic_container)

        if self._gptq:
            self._gptq_statistics = self._gptq_algo.get_statistic_points(
                model, graph, nodes_to_compress, self._backend_entity
            )
            statistics_aggregator.register_statistic_points(self._gptq_statistics)

        statistics_aggregator.collect_statistics(model, graph)

        for node_name, output_id in _collected_stat_inputs_map.items():
            act_node_name, output_port_id = output_id
            x_fp = self._get_fp_inputs(statistic_container, node_name=act_node_name, port_id=output_port_id)
            x_fp = [i.squeeze() for i in x_fp]  # List[tensor(seq_length, hidden_dim)]
            activations[node_name] = x_fp

            for shared_node_name in act_vs_shared_node_names_mapping[act_node_name]:
                activations[shared_node_name] = x_fp

        return activations



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.schemata.basic import ARRAY_OF_NUMBERS
from nncf.config.schemata.basic import BOOLEAN
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import STRING
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.defaults import RB_SPARSITY_SCHEDULER
from nncf.config.schemata.defaults import SPARSITY_FREEZE_EPOCH
from nncf.config.schemata.defaults import SPARSITY_LEVEL_SETTING_MODE
from nncf.config.schemata.defaults import SPARSITY_MULTISTEP_SPARSITY_LEVELS
from nncf.config.schemata.defaults import SPARSITY_MULTISTEP_STEPS
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_CONCAVE
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_PATIENCE
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_POWER
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_UPDATE_PER_OPTIMIZER_STEP
from nncf.config.schemata.defaults import SPARSITY_TARGET
from nncf.config.schemata.defaults import SPARSITY_TARGET_EPOCH

SCHEDULE_OPTIONS = [
    "polynomial",
    "exponential",
    "adaptive",
    "multistep",
]

COMMON_SPARSITY_PARAM_PROPERTIES = {
    "sparsity_level_setting_mode": with_attributes(
        STRING,
        description="The mode of sparsity level setting. \n"
        "`global` - the sparsity level is calculated across all "
        "weight values in the network across layers, "
        "`local` - the sparsity level can be set per-layer and "
        "within each layer is computed with respect only to the "
        "weight values within that layer.",
        default=SPARSITY_LEVEL_SETTING_MODE,
    ),
    "schedule": with_attributes(
        STRING,
        description=f"The type of scheduling to use for adjusting the target"
        f"sparsity level. Default - {RB_SPARSITY_SCHEDULER} for `rb_sparsity`, "
        f"{SPARSITY_SCHEDULER} otherwise",
        enum=SCHEDULE_OPTIONS,
    ),
    "sparsity_target": with_attributes(
        NUMBER,
        description="Target sparsity level for the model, to be reached at the end of the compression schedule.",
        default=SPARSITY_TARGET,
    ),
    "sparsity_target_epoch": with_attributes(
        NUMBER,
        description="Index of the epoch upon which the sparsity level "
        "of the model is scheduled to become "
        "equal to `sparsity_target`.",
        default=SPARSITY_TARGET_EPOCH,
    ),
    "sparsity_freeze_epoch": with_attributes(
        NUMBER,
        description="Index of the epoch upon which the sparsity mask will be frozen and no longer trained.",
        default=SPARSITY_FREEZE_EPOCH,
    ),
    "update_per_optimizer_step": with_attributes(
        BOOLEAN,
        description="Whether the function-based sparsity level schedulers "
        "should update the sparsity level after each optimizer "
        "step instead of each epoch step.",
        default=SPARSITY_SCHEDULER_UPDATE_PER_OPTIMIZER_STEP,
    ),
    "steps_per_epoch": with_attributes(
        NUMBER,
        description="Number of optimizer steps in one epoch. Required to start proper "
        "scheduling in the first training epoch if "
        "`update_per_optimizer_step` is `true.`",
    ),
    "multistep_steps": with_attributes(
        ARRAY_OF_NUMBERS,
        description="A list of scheduler steps at which to transition "
        "to the next scheduled sparsity level (multistep "
        "scheduler only).",
        default=SPARSITY_MULTISTEP_STEPS,
    ),
    "multistep_sparsity_levels": with_attributes(
        ARRAY_OF_NUMBERS,
        description="Multistep scheduler only - Levels of sparsity to use at "
        "each step of the scheduler as specified in the "
        "`multistep_steps` attribute. The first sparsity level "
        "will be applied immediately, "
        "so the length of this list should be larger than the "
        "length of the `multistep_steps` by one. The last "
        "sparsity level will function as the ultimate sparsity "
        'target, overriding the "sparsity_target" setting if it '
        "is present.",
        default=SPARSITY_MULTISTEP_SPARSITY_LEVELS,
    ),
    "patience": with_attributes(
        NUMBER,
        description="A conventional patience parameter for the scheduler, "
        "as for any other standard scheduler. Specified in units "
        "of scheduler steps.",
        default=SPARSITY_SCHEDULER_PATIENCE,
    ),
    "power": with_attributes(
        NUMBER,
        description="For polynomial scheduler - determines the corresponding power value.",
        default=SPARSITY_SCHEDULER_POWER,
    ),
    "concave": with_attributes(
        BOOLEAN,
        description="For polynomial scheduler - if `true`, then the target sparsity "
        "level will be approached in concave manner, and in convex "
        "manner otherwise.",
        default=SPARSITY_SCHEDULER_CONCAVE,
    ),
}
